import logging
from typing import List, Literal, Optional, Tuple

import numpy as np
import pyarrow as pa
import pyarrow.compute as pc
import quivr as qv

from ..constants import KM_P_AU
from ..constants import Constants as c
from ..coordinates import (
    CartesianCoordinates,
    GeodeticCoordinates,
    Origin,
    OriginCodes,
    SphericalCoordinates,
    transform_coordinates,
)
from ..orbits import Orbits, VariantOrbits
from ..propagator import Propagator
from ..time import Timestamp
from ..utils.plots.data import Coastlines
from ..utils.plots.logos import (
    AsteroidInstituteLogoDark,
    AsteroidInstituteLogoLight,
    get_logo_base64,
)
from ..utils.spice import get_perturber_state
from .impacts import CollisionEvent

EARTH_RADIUS_KM = c.R_EARTH_EQUATORIAL * KM_P_AU * 0.999
MOON_RADIUS_KM = 1738.1

logger = logging.getLogger(__name__)

try:
    import geopandas as gpd
    import plotly.graph_objects as go
except ImportError:
    raise ImportError("Please install adam_core[plots] to use this feature.")


def prepare_propagated_variants(
    propagated_variants: Orbits, impacts: CollisionEvent
) -> dict[Literal["Non-Impacting", "EARTH", "MOON"], Orbits]:
    """
    Sets variants propagated after their impact time to their impact coordinates on the surface of the colliding body. If the colliding body is the Earth,
    the variants are set to the impact coordinates on the surface of the Earth. If the colliding body is the Moon, the variants are set to the impact coordinates
    in the lunarcentric frame (but not fixed to the surface of the Moon).

    Note: Due to the nature of the impact detection code, some of the variants may already be inside the sphere of the Earth when the impact is detected.
    In these cases, the variants' distance from the geocenter is set to the radius of the Earth. The results of this function should not be used
    for high-fidelity impact predictions but instead for visualizations of the approximate impact corridor.

    Parameters
    ----------
    propagated_variants: Orbits
        The propagated variants to cleanse.
    impacts: CollisionEvent
        The impacts detected within the variants.

    Returns
    -------
    dict[Literal["Non-Impacting", "EARTH", "MOON"], Orbits]
        A dictionary containing the prepared variants, with keys:
          - "Non-Impacting": Variants that don't impact any body
          - "EARTH": Variants that impact Earth (if any)
          - "MOON": Variants that impact the Moon (if any)
        Only bodies that appear in the impacts will be included as keys.
    """
    assert propagated_variants.coordinates.frame == "ecliptic"

    colliding_bodies = impacts.collision_object.code.unique().to_pylist()
    prepared_variants = {}

    # Remove the non-impacting variants
    impacting_variants = propagated_variants.apply_mask(
        pc.is_in(propagated_variants.orbit_id, impacts.variant_id)
    )
    non_impacting_variants = propagated_variants.apply_mask(
        pc.invert(pc.is_in(propagated_variants.orbit_id, impacts.variant_id))
    )
    prepared_variants["Non-Impacting"] = non_impacting_variants

    for colliding_body in colliding_bodies:

        if colliding_body == "EARTH":
            radius = EARTH_RADIUS_KM
        elif colliding_body == "MOON":
            radius = MOON_RADIUS_KM
        else:
            raise ValueError(
                f"CollisionEvent visualizations are currently supported for the Earth and Moon. {colliding_body} is not supported."
            )

        impacts_on_colliding_body = impacts.apply_mask(
            pc.equal(impacts.collision_object.code, colliding_body)
        )
        impacting_variants_body = impacting_variants.apply_mask(
            pc.is_in(impacting_variants.orbit_id, impacts_on_colliding_body.variant_id)
        )

        for impact in impacts_on_colliding_body:

            post_impact_mask = pc.and_(
                pc.is_in(impacting_variants_body.orbit_id, impact.variant_id),
                pc.greater_equal(
                    impacting_variants_body.coordinates.time.mjd(),
                    impact.coordinates.time.rescale(
                        impacting_variants_body.coordinates.time.scale
                    ).mjd()[0],
                ),
            )

            impacting_variants_body_correct = impacting_variants_body.apply_mask(
                pc.invert(post_impact_mask)
            )
            impacting_variants_body_incorrect = impacting_variants_body.apply_mask(
                post_impact_mask
            )

            if len(impacting_variants_body_incorrect) > 0:
                collision_coordinates = impact.collision_coordinates
                collision_coordinates = qv.concatenate(
                    [
                        collision_coordinates
                        for _ in range(len(impacting_variants_body_incorrect))
                    ]
                )

                # Hack: In some cases, due to the time step of the propagation, the variants may already
                # be inside the sphere of the colliding body when the impact is detected. In these cases, the variants'
                # distance from the bodycenter is set its radius.
                collision_coordinates = collision_coordinates.set_column(
                    "rho", pa.repeat(radius / KM_P_AU, len(collision_coordinates))
                )

                # Override the time of the impact coordinates to the time of the propagated variants beyond the impact time, we do
                # this so we can then calculate position of these locations on the surface of the Earth as the Earth rotates.
                collision_coordinates = collision_coordinates.set_column(
                    "time", impacting_variants_body_incorrect.coordinates.time
                )

                geocentric_coordinates = transform_coordinates(
                    collision_coordinates,
                    representation_out=CartesianCoordinates,
                    frame_out="ecliptic",
                    origin_out=OriginCodes.EARTH,
                )
                impacting_variants_body_incorrect = (
                    impacting_variants_body_incorrect.set_column(
                        "coordinates",
                        geocentric_coordinates,
                    )
                )

            impacting_variants_body = qv.concatenate(
                [impacting_variants_body_correct, impacting_variants_body_incorrect]
            )

        prepared_variants[colliding_body] = impacting_variants_body

    return prepared_variants


def generate_impact_visualization_data(
    orbit: Orbits,
    variant_orbits: VariantOrbits,
    impacts: CollisionEvent,
    propagator: Propagator,
    time_step: float = 5,
    time_range: float = 60,
    max_processes: Optional[int] = None,
) -> Tuple[Timestamp, Orbits, dict[str, Orbits]]:
    """
    Generates the data for the impact visualization animation. The user should be careful
    to only send in collision events that correspond to an impact with a planetary body or moon.
    Non-impacting collisions such as close approaches have not been tested for this function.

    CollisionEvents visualizations are currently supported for the Earth and Moon.

    Parameters
    ----------
    orbit: Orbits
        The nominal best-fit orbit to propagate.
    variant_orbits: VariantOrbits
        The variants to propagate.
    impacts: CollisionEvent
        The impacts detected within the variants.
    propagator: Propagator
        The propagator to use to propagate the orbit.
    time_step: float
        The time step to use for the propagation.
    time_range: float
        The time range to use for the propagation.
    max_processes: Optional[int]
        The maximum number of processes to use for the propagation.

    Returns
    -------
    Tuple[Timestamp, Orbits, dict[str, Orbits]]
        The propagation times, the propagated nominal best-fit orbit and the propagated variants.
    """
    if pc.any(
        pc.invert(
            pc.or_(
                pc.equal(impacts.collision_object.code, "EARTH"),
                pc.equal(impacts.collision_object.code, "MOON"),
            )
        )
    ).as_py():
        raise ValueError(
            "CollisionEvents visualizations are currently supported for the Earth and Moon."
        )

    # Calculate the range of impact times
    impact_times = impacts.coordinates.time.mjd().to_numpy(zero_copy_only=False)
    first_impact_time = np.min(impact_times)
    last_impact_time = np.max(impact_times)

    # Create propagation times around the range of impact times
    mjds = np.arange(
        first_impact_time - time_range / 60 / 24,
        last_impact_time + time_range / 60 / 24 + time_step / 60 / 24,
        time_step / 60 / 24,
    )
    mjds = mjds - np.mod(mjds, time_step / 60 / 24)
    propagation_times = Timestamp.from_mjd(mjds, scale=impacts.coordinates.time.scale)

    # Propagate the variants to the propagation times
    propagated_variants = propagator.propagate_orbits(
        Orbits.from_kwargs(
            orbit_id=variant_orbits.variant_id,
            object_id=variant_orbits.object_id,
            coordinates=variant_orbits.coordinates,
        ),
        propagation_times,
        max_processes=max_processes,
    )
    # Transform the propagated variants to geocentric frame
    propagated_variants = propagated_variants.set_column(
        "coordinates",
        transform_coordinates(
            propagated_variants.coordinates,
            representation_out=CartesianCoordinates,
            frame_out="ecliptic",
            origin_out=OriginCodes.EARTH,
        ),
    )

    # Propagate the nominal best-fit orbit to the propagation times
    propagated_orbit = propagator.propagate_orbits(
        orbit, propagation_times, max_processes=max_processes
    )
    # Transform the best-fit orbit to geocentric frame
    propagated_orbit = propagated_orbit.set_column(
        "coordinates",
        transform_coordinates(
            propagated_orbit.coordinates,
            representation_out=CartesianCoordinates,
            frame_out="ecliptic",
            origin_out=OriginCodes.EARTH,
        ),
    )

    propagated_variants = prepare_propagated_variants(propagated_variants, impacts)

    for k, v in propagated_variants.items():
        propagated_variants[k] = v.sort_by(
            ["coordinates.time.days", "coordinates.time.nanos"]
        )

    return propagation_times, propagated_orbit, propagated_variants


def create_sphere(radius, offset=None):
    """
    Create a set of points that form a sphere.

    Parameters
    ----------
    radius: float
        The radius of the sphere.
    offset: array-like, optional
        The offset of the sphere from the origin.
    """
    if offset is None:
        offset = np.array([0, 0, 0])

    phi = np.linspace(0, np.pi, 50)
    theta = np.linspace(0, 2 * np.pi, 50)
    phi, theta = np.meshgrid(phi, theta)

    x = radius * np.sin(phi) * np.cos(theta)
    y = radius * np.sin(phi) * np.sin(theta)
    z = radius * np.cos(phi)

    x += offset[0]
    y += offset[1]
    z += offset[2]

    return x, y, z


def add_earth(
    time,
    coastlines: bool = True,
    origin: OriginCodes = OriginCodes.EARTH,
    frame: str = "ecliptic",
    show: bool = True,
) -> Tuple[go.Surface, List[go.Scatter3d]]:
    """
    Add the Earth to the plot.

    Parameters
    ----------
    time: Timestamp
        The time of the plot.
    coastlines: bool, optional
        Whether to add the coastlines to the plot.
    origin: Origin, optional
        The origin of the plot.
    frame: str, optional
        The frame of the plot.
    show: bool, optional
        Whether to show the Earth by default.

    Returns
    -------
    Tuple[go.Surface, List[go.Scatter3d]]
        The Earth rendered as a sphere and the traces for the coastlines.
    """
    earth_state = get_perturber_state(
        OriginCodes.EARTH,
        time,
        frame=frame,
        origin=origin,
    )
    x, y, z = create_sphere(EARTH_RADIUS_KM * 0.999, offset=earth_state.r[0] * KM_P_AU)

    surface_traces = []
    if coastlines:
        world = gpd.read_file(str(Coastlines))

        for idx, row in world.iterrows():
            # Get the polygon coordinates
            if row.geometry.geom_type == "Polygon":
                coords = np.array(row.geometry.exterior.coords)
            elif row.geometry.geom_type == "MultiPolygon":
                coords = np.array(row.geometry.geoms[0].exterior.coords)
            elif row.geometry.geom_type == "LineString":
                coords = np.array(row.geometry.coords)
            elif row.geometry.geom_type == "MultiLineString":
                coords = np.array(row.geometry.geoms[0].coords)

            coords = SphericalCoordinates.from_kwargs(
                rho=pa.repeat(EARTH_RADIUS_KM / KM_P_AU, len(coords)),
                lon=coords[:, 0],
                lat=coords[:, 1],
                time=Timestamp.from_kwargs(
                    days=pa.repeat(time.days[0], len(coords)),
                    nanos=pa.repeat(time.nanos[0], len(coords)),
                    scale=time.scale,
                ),
                frame="itrf93",
                origin=Origin.from_kwargs(
                    code=pa.repeat(OriginCodes.EARTH.name, len(coords))
                ),
            )

            coords = transform_coordinates(
                coords,
                representation_out=CartesianCoordinates,
                frame_out=frame,
                origin_out=origin,
            )

            surface_traces.append(
                go.Scatter3d(
                    x=coords.x.to_numpy(zero_copy_only=False) * KM_P_AU,
                    y=coords.y.to_numpy(zero_copy_only=False) * KM_P_AU,
                    z=coords.z.to_numpy(zero_copy_only=False) * KM_P_AU,
                    mode="lines",
                    line=dict(color="white", width=2),
                    showlegend=False,
                    visible=show,
                    legendgroup="Earth",
                )
            )

    earth_surface = go.Surface(
        x=x,
        y=y,
        z=z,
        opacity=1,
        colorscale=[[0, "#015294"], [1, "#015294"]],
        showscale=False,
        name="Earth",
        legendgroup="Earth",
        showlegend=True,
        visible=show,
    )
    return earth_surface, surface_traces


def add_moon(
    time: Timestamp,
    origin: OriginCodes = OriginCodes.EARTH,
    frame: Literal["ecliptic", "equatorial", "itrf93"] = "ecliptic",
    show: bool = True,
) -> go.Surface:
    """
    Add the Moon to the plot.

    Parameters
    ----------
    time: Timestamp
        The time of the snapshot.
    origin: OriginCodes
        The origin of the plot.
    frame: Literal["ecliptic", "equatorial", "itrf93"]
        The frame of the plot.

    Returns
    -------
    go.Surface
        The Moon rendered as a sphere.
    """
    lunar_state = get_perturber_state(
        OriginCodes.MOON,
        time,
        frame=frame,
        origin=origin,
    )
    x, y, z = create_sphere(MOON_RADIUS_KM, offset=lunar_state.r[0] * KM_P_AU)

    return go.Surface(
        x=x,
        y=y,
        z=z,
        opacity=1.0,
        colorscale=[[0, "#A9A9A9"], [1, "#A9A9A9"]],
        showscale=False,
        name="Moon",
        showlegend=True,
        visible=show,
    )


def plot_impact_simulation(
    propagation_times: Timestamp,
    propagated_best_fit_orbit: Orbits,
    propagated_variants: dict[str, Orbits],
    impacts: CollisionEvent,
    grid: bool = True,
    title: str = None,
    logo: bool = True,
    show_impacting: bool = True,
    show_non_impacting: bool = True,
    show_best_fit: bool = True,
    show_earth: bool = True,
    show_moon: bool = True,
    sample_impactors: Optional[float] = None,
    sample_non_impactors: Optional[float] = None,
    height: Optional[int] = None,
    width: Optional[int] = None,
) -> go.Figure:
    """
    Plot the impact simulation.

    Parameters
    ----------
    propagated_best_fit_orbit: Orbits
        The propagated best-fit orbit.
    propagated_variants: Orbits
        The propagated variants.
    impacts: CollisionEvent
        The impacts detected within the variants.
    grid: bool, optional
        Whether to add the grid to the plot.
    title: str, optional
        The title of the plot.
    logo: bool, optional
        Whether to add the Asteroid Institute logo to the plot.
    show_impacting: bool, optional
        Whether to show the impacting variants.
    show_non_impacting: bool, optional
        Whether to show the non-impacting variants.
    show_best_fit: bool, optional
        Whether to show the best-fit orbit.
    show_earth: bool, optional
        Whether to show the Earth.
    show_moon: bool, optional
        Whether to show the Moon.
    sample_impactors: Optional[float], optional
        Randomly sample the impactors for plotting. Should be between 0 and 1.
    sample_non_impactors: Optional[float], optional
        Randomly sample the non-impactors for plotting. Should be between 0 and 1.
    height: int, optional
        The height of the plot.
    width: int, optional
        The width of the plot.

    Returns
    -------
    go.Figure
        The impact simulation plot.
    """
    propagation_times_isot = propagation_times.to_astropy().isot

    num_variants = 0
    impact_count = {}
    sampled_variants = {}
    for k, v in propagated_variants.items():
        num_variants += len(v.orbit_id.unique())

        if k == "Non-Impacting":
            if sample_non_impactors is not None:
                orbit_ids = v.orbit_id.unique()
                numpy_orbit_ids = orbit_ids.to_numpy(
                    zero_copy_only=False
                )  # Convert to NumPy

                if len(numpy_orbit_ids) == 0:
                    orbit_ids_sample = numpy_orbit_ids  # Already an empty numpy array with correct dtype
                    logger.info("No non-impacting variants available to sample.")
                else:
                    sample_size = np.ceil(
                        len(numpy_orbit_ids) * sample_non_impactors
                    ).astype(int)
                    sample_size = min(
                        sample_size, len(numpy_orbit_ids)
                    )  # Ensure sample_size <= population
                    orbit_ids_sample = np.random.choice(
                        numpy_orbit_ids,
                        sample_size,
                        replace=False,
                    )
                    logger.info(
                        f"Sampled {len(orbit_ids_sample)} non-impacting variants out of {len(numpy_orbit_ids)}"
                    )

                # Create Arrow array with explicit type
                arrow_orbit_ids_sample = pa.array(
                    orbit_ids_sample, type=v.orbit_id.type
                )
                sampled_variants[k] = v.__class__.from_pyarrow(
                    v.apply_mask(
                        pc.is_in(v.orbit_id, arrow_orbit_ids_sample)
                    ).table.combine_chunks()
                )
            else:
                sampled_variants[k] = v

        if k != "Non-Impacting":
            impact_count[k] = 0

            if sample_impactors is not None:
                orbit_ids = v.orbit_id.unique()
                numpy_orbit_ids = orbit_ids.to_numpy(
                    zero_copy_only=False
                )  # Convert to NumPy

                if len(numpy_orbit_ids) == 0:
                    orbit_ids_sample = numpy_orbit_ids  # Empty array with correct dtype
                    logger.info(f"No impacting variants for '{k}' available to sample.")
                else:
                    sample_size = np.ceil(
                        len(numpy_orbit_ids) * sample_impactors
                    ).astype(int)
                    sample_size = min(
                        sample_size, len(numpy_orbit_ids)
                    )  # Ensure sample_size <= population
                    orbit_ids_sample = np.random.choice(
                        numpy_orbit_ids,
                        sample_size,
                        replace=False,
                    )
                    logger.info(
                        f"Sampled {len(orbit_ids_sample)} impacting variants for '{k}' out of {len(numpy_orbit_ids)}"
                    )

                # Create Arrow array with explicit type
                arrow_orbit_ids_sample = pa.array(
                    orbit_ids_sample, type=v.orbit_id.type
                )
                sampled_variants[k] = v.__class__.from_pyarrow(
                    v.apply_mask(
                        pc.is_in(v.orbit_id, arrow_orbit_ids_sample)
                    ).table.combine_chunks()
                )
            else:
                sampled_variants[k] = v

    all_potential_impactor_ids_from_impacts_table = set(
        impacts.variant_id.unique().to_pylist()
    )

    if title is None:
        prefix = ""
    else:
        prefix = f"{title}<br>"

    # Build the individual frames for the animation
    frames = []
    for i, time in enumerate(propagation_times):

        # 1. Get all impacts up to the current time
        all_impacts_up_to_current_time = impacts.apply_mask(
            pc.less_equal(impacts.coordinates.time.mjd(), time.mjd()[0])
        )

        # 2. Get the set of unique variant IDs that have impacted *anything* up to current time
        current_frame_total_unique_impacted_ids_set = set(
            all_impacts_up_to_current_time.variant_id.unique().to_pylist()
        )

        # 3. Update impact_count for each body (for title text)
        for (
            body_key
        ) in (
            impact_count.keys()
        ):  # These are "EARTH", "MOON", etc. as initialized earlier
            impacts_on_this_body_up_to_current_time = (
                all_impacts_up_to_current_time.apply_mask(
                    pc.equal(
                        all_impacts_up_to_current_time.collision_object.code, body_key
                    )
                )
            )
            impact_count[body_key] = len(
                impacts_on_this_body_up_to_current_time.variant_id.unique()
            )

        # Create the data for the frame
        data = []
        for k, v in sampled_variants.items():

            if k == "Non-Impacting" and not show_non_impacting:
                continue

            if k == "Impacting" and not show_impacting:
                continue

            v_at_time = v.apply_mask(
                pc.and_(
                    pc.equal(v.coordinates.time.days, time.days[0]),
                    pc.and_(
                        pc.less_equal(
                            v.coordinates.time.nanos, time.nanos[0].as_py() + 100000
                        ),
                        pc.greater_equal(
                            v.coordinates.time.nanos, time.nanos[0].as_py() - 100000
                        ),
                    ),
                )
            )

            if k == "Non-Impacting":
                color = "#5685C3"
                size = 1
                name = k
            else:
                color = "red"
                size = 2
                name = f"{k.lower().capitalize()} Impacting"

            x = v_at_time.coordinates.x.to_numpy(zero_copy_only=False) * KM_P_AU
            y = v_at_time.coordinates.y.to_numpy(zero_copy_only=False) * KM_P_AU
            z = v_at_time.coordinates.z.to_numpy(zero_copy_only=False) * KM_P_AU

            data.append(
                go.Scatter3d(
                    x=x,
                    y=y,
                    z=z,
                    mode="markers",
                    marker=dict(
                        size=size,
                        color=color,
                        opacity=1,
                        showscale=False,
                    ),
                    name=name,
                    visible=True,
                    showlegend=True,
                )
            )

        # Add best-fit orbit
        if show_best_fit:
            orbit_at_time = propagated_best_fit_orbit[i].coordinates
            data.append(
                go.Scatter3d(
                    x=orbit_at_time.x.to_numpy(zero_copy_only=False) * KM_P_AU,
                    y=orbit_at_time.y.to_numpy(zero_copy_only=False) * KM_P_AU,
                    z=orbit_at_time.z.to_numpy(zero_copy_only=False) * KM_P_AU,
                    mode="markers",
                    marker=dict(
                        size=3,
                        color="#F07620",
                        opacity=1,
                        showscale=False,
                    ),
                    name="Best-Fit Orbit",
                    visible=True,
                    showlegend=True,
                )
            )

        earth_surface, surface_traces = add_earth(time, show=show_earth)
        data.append(earth_surface)
        data.extend(surface_traces)
        data.append(add_moon(time, show=show_moon))

        text = f"{prefix}Time: {propagation_times_isot[i]}"
        for k in impact_count:
            text += f"<br>{k.lower().capitalize()} Impacts: {impact_count[k]} of {num_variants} Variants<br>{k.lower().capitalize()} Impact Probability: {impact_count[k]/num_variants * 100:.3f}%"

        frame = go.Frame(
            data=data,
            name=str(i),
            layout=dict(
                title=dict(
                    text=text,
                    x=0.01,
                    y=0.97,
                    font=dict(size=14, color="white"),
                ),
            ),
        )

        frames.append(frame)

        # If all impacting variants (from the original impacts table) have impacted by this frame, stop.
        if all_potential_impactor_ids_from_impacts_table:
            if all_potential_impactor_ids_from_impacts_table.issubset(
                current_frame_total_unique_impacted_ids_set
            ):
                break

    # Plot the figure
    fig = go.Figure(data=frames[0].data, frames=frames, layout=frames[0].layout)

    if grid:
        config = dict(
            showgrid=True,
            zeroline=True,
            visible=True,
            showticklabels=True,
            gridcolor="rgba(128,128,128,0.1)",
            showbackground=False,
            backgroundcolor="rgb(0,0,0)",
            tickfont=dict(color="white"),
        )
    else:
        config = dict(
            showgrid=False,
            zeroline=False,
            visible=False,
            showticklabels=False,
            gridcolor="rgba(128,128,128,0.0)",
            showbackground=False,
            backgroundcolor="rgb(0,0,0)",
            tickfont=dict(color="white"),
        )

    if logo:
        images = [
            dict(
                source=get_logo_base64(AsteroidInstituteLogoDark),
                xref="paper",
                yref="paper",
                x=0.96,
                y=-0.15,
                sizex=0.20,
                sizey=0.20,
                xanchor="left",
                yanchor="bottom",
                layer="above",
            )
        ]
    else:
        images = []

    fig.update_layout(
        scene=dict(
            aspectratio=dict(x=1, y=1, z=1),
            aspectmode="data",
            camera=dict(
                center=dict(x=0, y=0, z=0),
                eye=dict(x=1.5, y=1.5, z=1.5),
                up=dict(x=0, y=0, z=1),
            ),
            bgcolor="rgb(0,0,0)",
            xaxis=dict(title=dict(text="x [km]", font=dict(color="white")), **config),
            yaxis=dict(title=dict(text="y [km]", font=dict(color="white")), **config),
            zaxis=dict(title=dict(text="z [km]", font=dict(color="white")), **config),
        ),
        height=height,
        width=width,
        autosize=True,
        margin=dict(l=7, r=7, t=10, b=7, pad=0),
        paper_bgcolor="rgb(0,0,0)",
        plot_bgcolor="rgb(0,0,0)",
        font=dict(color="white"),
        images=images,
        updatemenus=[
            dict(
                type="buttons",
                showactive=False,
                x=0.03,
                y=-0.03,
                buttons=[
                    dict(
                        label="▶",
                        method="animate",
                        args=[
                            None,
                            {
                                "frame": {"duration": 50, "redraw": True},
                                "fromcurrent": True,
                            },
                        ],
                    ),
                    dict(
                        label="⏸",
                        method="animate",
                        args=[
                            [None],
                            {
                                "frame": {"duration": 0, "redraw": False},
                                "mode": "immediate",
                                "transition": {"duration": 0},
                            },
                        ],
                    ),
                ],
            )
        ],
        sliders=[
            dict(
                currentvalue=dict(prefix="Time: "),
                pad=dict(t=50),
                len=0.90 if not logo else 0.80,
                x=0.10,
                y=0.0,
                font=dict(color="white", size=10),
                steps=[
                    dict(
                        args=[
                            [str(i)],
                            dict(
                                frame=dict(duration=50, redraw=True),
                                mode="immediate",
                            ),
                        ],
                        label=propagation_times_isot[i],
                        method="animate",
                    )
                    for i in range(len(frames))
                ],
            )
        ],
        uirevision="constant",
    )

    return fig


def plot_risk_corridor(
    impacts: CollisionEvent,
    title: Optional[str] = None,
    logo: bool = True,
    height: Optional[int] = None,
    width: Optional[int] = None,
) -> go.Figure:
    """
    Plot the risk corridor with toggleable globe/map views.
    Points colored by time with a linear scale and animated sequence.

    Parameters
    ----------
    impacts : Impact data containing coordinates
    title : str, optional
        Plot title
    logo : bool, optional
        Whether to add the Asteroid Institute logo to the plot.
    height : int, optional
        The height of the plot.
    width : int, optional
        The width of the plot.

    Returns
    -------
    go.Figure
        The risk corridor plot.
    """
    # Filter to only include impacts on the Earth
    impacts = impacts.apply_mask(pc.equal(impacts.collision_object.code, "EARTH"))
    if len(impacts) == 0:
        raise ValueError(
            "No Earth impacts found. Other collision objects are not supported yet."
        )

    # Transform impact coordinates to ITRF93 Geodetic Coordinates
    geodetic_impacts = transform_coordinates(
        impacts.collision_coordinates,
        representation_out=GeodeticCoordinates,
        frame_out="itrf93",
        origin_out=OriginCodes.EARTH,
    )

    # Sort all data by time
    times = geodetic_impacts.time.to_astropy()
    time_order = np.argsort(times.mjd)
    lon = geodetic_impacts.lon.to_numpy(zero_copy_only=False)[time_order]
    lat = geodetic_impacts.lat.to_numpy(zero_copy_only=False)[time_order]
    times = times[time_order]

    # Convert times to minutes since first impact
    time_nums = (times.mjd - times.mjd.min()) * 24 * 60

    # Calculate center
    center_lon = lon[0]
    center_lat = lat[0]

    plot_config = dict(
        height=height,
        width=width,
        map=dict(
            style="open-street-map",
            center=dict(lat=center_lat, lon=center_lon),
            zoom=1,
        ),
    )
    color_bar_config = dict(
        tickmode="array",
        tickangle=0,
        orientation="v",  # Vertical orientation for right side
        x=1.02,
        y=0.5,  # Centered vertically
        xanchor="left",
        yanchor="middle",
        thickness=25,
        len=0.8,
    )

    # Create frames for animation
    frames = []
    for i in range(len(lon)):
        # Create ticks for min, middle (when applicable), and max
        current_max_time = time_nums[i]

        if i == 0:
            # First frame: only show T+0 min
            tick_values = [0]
            tick_labels = ["T+0 min"]
        elif i == 1:
            # Second frame: show T+0 min and max
            tick_values = [0, current_max_time]
            tick_labels = ["T+0 min", f"T+{current_max_time:.0f} min"]
        else:
            # Third frame and beyond: show T+0 min, middle, and max
            middle_time = current_max_time / 2
            tick_values = [0, middle_time, current_max_time]
            tick_labels = [
                "T+0 min",
                f"T+{middle_time:.0f} min",
                f"T+{current_max_time:.0f} min",
            ]
        frame = go.Frame(
            data=[
                go.Scattermap(
                    lon=lon[: i + 1],
                    lat=lat[: i + 1],
                    mode="markers",
                    marker=dict(
                        size=8,
                        color=time_nums[: i + 1],
                        colorscale="Viridis",
                        opacity=0.8,
                        showscale=True,
                        colorbar=dict(
                            title=dict(
                                side="right",
                                font=dict(size=12, color="black"),
                            ),
                            ticktext=tick_labels,
                            tickvals=tick_values,
                            **color_bar_config,
                        ),
                    ),
                    name="Variant Impact Locations",
                    hovertext=[
                        f"Time: {t.iso}<br>Lon: {lo:.2f}°<br>Lat: {la:.2f}°<br>+{mins:.1f} min"
                        for t, lo, la, mins in zip(
                            times[: i + 1],
                            lon[: i + 1],
                            lat[: i + 1],
                            time_nums[: i + 1],
                        )
                    ],
                )
            ],
            name=str(i),
        )
        frames.append(frame)

    # Create the figure with initial state
    fig = go.Figure(
        data=[
            go.Scattermap(
                lon=[np.nan],
                lat=[np.nan],
                mode="markers",
                marker=dict(
                    size=8,
                    color=[],
                    colorscale="Viridis",
                    opacity=0.8,
                    showscale=True,
                    colorbar=dict(
                        title=dict(
                            side="right",
                            font=dict(size=12, color="black"),
                        ),
                        ticktext=["T+0 min", f"T+{time_nums[-1]:.0f} min"],
                        tickvals=[0, time_nums[-1]],
                        **color_bar_config,
                    ),
                ),
                name="Variant Impact Locations",
                hovertext=[
                    f"Time: {times[0].iso}<br>Lon: {lon[0]:.2f}°<br>Lat: {lat[0]:.2f}°<br>+{time_nums[0]:.1f} min"
                ],
            )
        ],
        frames=frames,
    )

    if title is None:
        title = "Risk Corridor"

    if logo:
        images = [
            dict(
                source=get_logo_base64(AsteroidInstituteLogoLight),
                xref="paper",
                yref="paper",
                x=0.81,
                y=-0.20,  # Moved logo up slightly
                sizex=0.18,
                sizey=0.18,
                xanchor="left",
                yanchor="bottom",
                layer="above",
            )
        ]
    else:
        images = []

    # Update layout
    fig.update_layout(
        title=dict(
            text=title,
            x=0.01,
            y=0.98,
            font=dict(size=14, color="black"),
        ),
        autosize=True,
        margin=dict(
            l=7, r=120, t=30, b=30, pad=0
        ),  # Increased right margin for vertical colorbar
        **plot_config,
        updatemenus=[
            dict(
                type="buttons",
                showactive=False,
                x=0.03,
                y=-0.12,  # Positioned between slider and colorbar
                buttons=[
                    dict(
                        label="▶",
                        method="animate",
                        args=[
                            None,
                            {
                                "frame": {"duration": 50, "redraw": True},
                                "fromcurrent": True,
                            },
                        ],
                    ),
                    dict(
                        label="⏸",
                        method="animate",
                        args=[
                            [None],
                            {
                                "frame": {"duration": 0, "redraw": False},
                                "mode": "immediate",
                                "transition": {"duration": 0},
                            },
                        ],
                    ),
                ],
            )
        ],
        images=images,
        sliders=[
            dict(
                currentvalue=dict(prefix="Variant: "),
                pad=dict(t=50),
                len=0.90 if not logo else 0.75,
                x=0.05,
                font=dict(color="black", size=10),
                steps=[
                    dict(
                        args=[
                            [str(i)],
                            dict(
                                frame=dict(duration=50, redraw=True),
                                mode="immediate",
                                transition=dict(duration=0),
                            ),
                        ],
                        label=str(i),
                        method="animate",
                    )
                    for i in range(len(frames))
                ],
            )
        ],
    )

    return fig



import logging
import multiprocessing as mp
import time
from itertools import combinations
from typing import Literal, Optional, Tuple, Type, Union

import numpy as np
import numpy.typing as npt
import pyarrow as pa
import pyarrow.compute as pc
import quivr as qv
import ray

from ..coordinates.residuals import Residuals
from ..propagator import Propagator
from ..ray_cluster import initialize_use_ray
from ..utils.iter import _iterate_chunk_indices, _iterate_chunks
from . import (
    FittedOrbitMembers,
    FittedOrbits,
    OrbitDeterminationObservations,
    drop_duplicate_orbits,
)
from .gauss import gaussIOD
from .outliers import calculate_max_outliers

logger = logging.getLogger(__name__)

__all__ = ["initial_orbit_determination"]


def sort_by_id_and_time(
    linkages: qv.AnyTable,
    members: qv.AnyTable,
    observations: OrbitDeterminationObservations,
    linkage_column: str,
) -> Tuple[qv.AnyTable, qv.AnyTable]:
    """
    Sort linkages and linkage members by linkage ID and observation time.

    Parameters
    ----------
    linkages : qv.AnyTable
        Linkages to sort.
    members : qv.AnyTable
        Linkage members to sort.
    observations : Observations
        Observations from which linkage members were generated. Observations
        are used to determine the observation time of each linkage member.
    linkage_column : str
        Column name in the linkage table to use for sorting. For clusters
        this is "cluster_id" and for orbits this is "orbit_id".

    Returns
    -------
    linkages : qv.AnyTable
        Sorted linkages.
    members : qv.AnyTable
        Sorted linkage members.
    """
    # Grab the linkage ID column from the linkages table and add an index column
    linkage_table = linkages.table.select([linkage_column])
    linkage_table = linkage_table.add_column(
        0, "index", pa.array(np.arange(0, len(linkage_table)))
    )

    # Grab the linkage ID and observation ID columns from the linkage members table and add an index column
    members_table = members.table.select([linkage_column, "obs_id"])
    members_table = members_table.add_column(
        0, "index", pa.array(np.arange(0, len(members_table)))
    )

    # Grab the observation ID, observation time columns and join with the linkage members table on the observation ID
    observation_times = observations.flattened_table().select(
        ["id", "coordinates.time.days", "coordinates.time.nanos"]
    )
    member_times = members_table.join(
        observation_times, keys=["obs_id"], right_keys=["id"]
    )

    # Sort the reduced linkages table by linkage ID and the linkage member times table by linkage ID and observation time
    linkage_table = linkage_table.sort_by([(linkage_column, "ascending")])
    member_times = member_times.sort_by(
        [
            (linkage_column, "ascending"),
            ("coordinates.time.days", "ascending"),
            ("coordinates.time.nanos", "ascending"),
        ]
    )

    linkages = linkages.take(linkage_table["index"])
    members = members.take(member_times["index"])

    if linkages.fragmented():
        linkages = qv.defragment(linkages)
    if members.fragmented():
        members = qv.defragment(members)
    return linkages, members


def select_observations(
    observations: OrbitDeterminationObservations,
    method: Literal["combinations", "first+middle+last", "thirds"] = "combinations",
) -> npt.NDArray[np.str_]:
    """
    Selects which three observations to use for IOD depending on the method.

    Methods:
        'first+middle+last' : Grab the first, middle and last observations in time.
        'thirds' : Grab the middle observation in the first third, second third, and final third.
        'combinations' : Return the observation IDs corresponding to every possible combination of three observations with
            non-coinciding observation times.

    Parameters
    ----------
    observations : `~pandas.DataFrame`
        Pandas DataFrame containing observations with at least a column of observation IDs and a column
        of exposure times.
    method : {'first+middle+last', 'thirds', 'combinations'}, optional
        Which method to use to select observations.
        [Default = 'combinations']

    Returns
    -------
    obs_id : `~numpy.ndarray' (N, 3 or 0)
        An array of selected observation IDs. If three unique observations could
        not be selected then returns an empty array.
    """
    obs_ids = observations.id.to_numpy(zero_copy_only=False)
    if len(obs_ids) < 3:
        return np.array([])

    indexes = np.arange(0, len(obs_ids))
    times = observations.coordinates.time.mjd().to_numpy(zero_copy_only=False)

    if method == "first+middle+last":
        selected_times = np.percentile(times, [0, 50, 100], interpolation="nearest")
        selected_index = np.intersect1d(times, selected_times, return_indices=True)[1]
        selected_index = np.array([selected_index])

    elif method == "thirds":
        selected_times = np.percentile(
            times, [1 / 6 * 100, 50, 5 / 6 * 100], interpolation="nearest"
        )
        selected_index = np.intersect1d(times, selected_times, return_indices=True)[1]
        selected_index = np.array([selected_index])

    elif method == "combinations":
        # Make all possible combinations of 3 observations
        selected_index = np.array(
            [np.array(index) for index in combinations(indexes, 3)]
        )

        # Calculate arc length
        arc_length = times[selected_index][:, 2] - times[selected_index][:, 0]

        # Calculate distance of second observation from middle point (last + first) / 2
        time_from_mid = np.abs(
            (times[selected_index][:, 2] + times[selected_index][:, 0]) / 2
            - times[selected_index][:, 1]
        )

        # Sort by descending arc length and ascending time from midpoint
        sort = np.lexsort((time_from_mid, -arc_length))
        selected_index = selected_index[sort]

    else:
        raise ValueError("method should be one of {'first+middle+last', 'thirds'}")

    # Make sure each returned combination of observation ids have at least 3 unique
    # times
    keep = []
    for i, comb in enumerate(times[selected_index]):
        if len(np.unique(comb)) == 3:
            keep.append(i)
    keep = np.array(keep)

    # Return an empty array if no observations satisfy the criteria
    if len(keep) == 0:
        return np.array([])
    else:
        selected_index = selected_index[keep, :]

    return obs_ids[selected_index]


def iod_worker(
    linkage_ids: npt.NDArray[np.str_],
    observations: Union[OrbitDeterminationObservations, ray.ObjectRef],
    linkage_members: Union[FittedOrbitMembers, ray.ObjectRef],
    propagator: Type[Propagator],
    min_obs: int = 6,
    min_arc_length: float = 1.0,
    contamination_percentage: float = 0.0,
    rchi2_threshold: float = 200,
    observation_selection_method: Literal[
        "combinations", "first+middle+last", "thirds"
    ] = "combinations",
    linkage_id_col: str = "cluster_id",
    iterate: bool = False,
    light_time: bool = True,
    propagator_kwargs: dict = {},
) -> Tuple[FittedOrbits, FittedOrbitMembers]:

    iod_orbits = FittedOrbits.empty()
    iod_orbit_members = FittedOrbitMembers.empty()
    for linkage_id in linkage_ids:
        time_start = time.time()
        logger.debug(f"Finding initial orbit for linkage {linkage_id}...")

        obs_ids = linkage_members.apply_mask(
            pc.equal(linkage_members.column(linkage_id_col), linkage_id)
        ).obs_id
        observations_linkage = observations.apply_mask(
            pc.is_in(observations.id, obs_ids)
        )

        # Sort observations by time
        observations_linkage = observations_linkage.sort_by(
            [
                "coordinates.time.days",
                "coordinates.time.nanos",
                "coordinates.origin.code",
            ]
        )

        iod_orbit, iod_orbit_orbit_members = iod(
            observations_linkage,
            min_obs=min_obs,
            min_arc_length=min_arc_length,
            rchi2_threshold=rchi2_threshold,
            contamination_percentage=contamination_percentage,
            observation_selection_method=observation_selection_method,
            iterate=iterate,
            light_time=light_time,
            propagator=propagator,
            propagator_kwargs=propagator_kwargs,
        )
        if len(iod_orbit) > 0:
            iod_orbit = iod_orbit.set_column("orbit_id", pa.array([linkage_id]))
            iod_orbit_orbit_members = iod_orbit_orbit_members.set_column(
                "orbit_id",
                pa.array([linkage_id for i in range(len(iod_orbit_orbit_members))]),
            )

        time_end = time.time()
        duration = time_end - time_start
        logger.debug(f"IOD for linkage {linkage_id} completed in {duration:.3f}s.")

        iod_orbits = qv.concatenate([iod_orbits, iod_orbit])
        if iod_orbits.fragmented():
            iod_orbits = qv.defragment(iod_orbits)

        iod_orbit_members = qv.concatenate([iod_orbit_members, iod_orbit_orbit_members])
        if iod_orbit_members.fragmented():
            iod_orbit_members = qv.defragment(iod_orbit_members)

    return iod_orbits, iod_orbit_members


@ray.remote
def iod_worker_remote(
    linkage_ids: Union[npt.NDArray[np.str_], ray.ObjectRef],
    linkage_members_indices: Tuple[int, int],
    observations: Union[OrbitDeterminationObservations, ray.ObjectRef],
    linkage_members: Union[FittedOrbitMembers, ray.ObjectRef],
    propagator: Type[Propagator],
    min_obs: int = 6,
    min_arc_length: float = 1.0,
    contamination_percentage: float = 0.0,
    rchi2_threshold: float = 200,
    observation_selection_method: Literal[
        "combinations", "first+middle+last", "thirds"
    ] = "combinations",
    linkage_id_col: str = "cluster_id",
    iterate: bool = False,
    light_time: bool = True,
    propagator_kwargs: dict = {},
) -> Tuple[FittedOrbits, FittedOrbitMembers]:
    # Select linkage ids from linkage_members_indices
    linkage_id_chunk = linkage_ids[
        linkage_members_indices[0] : linkage_members_indices[1]
    ]
    return iod_worker(
        linkage_id_chunk,
        observations,
        linkage_members,
        min_obs=min_obs,
        min_arc_length=min_arc_length,
        contamination_percentage=contamination_percentage,
        rchi2_threshold=rchi2_threshold,
        observation_selection_method=observation_selection_method,
        linkage_id_col=linkage_id_col,
        iterate=iterate,
        light_time=light_time,
        propagator=propagator,
        propagator_kwargs=propagator_kwargs,
    )


iod_worker_remote.options(num_returns=1, num_cpus=1)


def iod(
    observations: OrbitDeterminationObservations,
    propagator: Type[Propagator],
    min_obs: int = 6,
    min_arc_length: float = 1.0,
    contamination_percentage: float = 0.0,
    rchi2_threshold: float = 200,
    observation_selection_method: Literal[
        "combinations", "first+middle+last", "thirds"
    ] = "combinations",
    iterate: bool = False,
    light_time: bool = True,
    propagator_kwargs: dict = {},
) -> Tuple[FittedOrbits, FittedOrbitMembers]:
    """
    Run initial orbit determination on a set of observations believed to belong to a single
    object.

    Parameters
    ----------
    observations : `OrbitDeterminationObservations`
        Observations with at least the following attributes:
            id : Observation IDs [str],
            coordinates : SkyCoord object containing observation time, RA, and Dec
            observers : Observer object containing observatory positions
    min_obs : int, optional
        Minimum number of observations that must remain in the linkage. For example, if min_obs is set to 6 and
        a linkage has 8 observations, at most the two worst observations will be flagged as outliers if their individual
        chi2 values exceed the chi2 threshold.
    contamination_percentage : float, optional
        Maximum percent of observations that can flagged as outliers.
    rchi2_threshold : float, optional
        Maximum reduced chi2 required for an initial orbit to be accepted.
    observation_selection_method : {'first+middle+last', 'thirds', 'combinations'}, optional
        Selects which three observations to use for IOD depending on the method. The avaliable methods are:
            'first+middle+last' : Grab the first, middle and last observations in time.
            'thirds' : Grab the middle observation in the first third, second third, and final third.
            'combinations' : Return the observation IDs corresponding to every possible combination of three observations with
                non-coinciding observation times.
    iterate : bool, optional
        Iterate the preliminary orbit solution using the state transition iterator.
    light_time : bool, optional
        Correct preliminary orbit for light travel time.
    propagator : Type[Propagator], optional
        Which propagator to use for ephemeris generation.
    propagator_kwargs : dict, optional
        Settings and additional parameters to pass to selected
        propagator.

    Returns
    -------
    iod_orbits : `FittedOrbits`
        Dataframe with orbits found in linkages.
            "orbit_id" : Orbit ID, a uuid [str],
            "epoch" : Epoch at which orbit is defined in MJD TDB [float],
            "x" : Orbit's ecliptic J2000 x-position in au [float],
            "y" : Orbit's ecliptic J2000 y-position in au [float],
            "z" : Orbit's ecliptic J2000 z-position in au [float],
            "vx" : Orbit's ecliptic J2000 x-velocity in au per day [float],
            "vy" : Orbit's ecliptic J2000 y-velocity in au per day [float],
            "vz" : Orbit's ecliptic J2000 z-velocity in au per day [float],
            "arc_length" : Arc length in days [float],
            "num_obs" : Number of observations that were within the chi2 threshold
                of the orbit.
            "chi2" : Total chi2 of the orbit calculated using the predicted location of the orbit
                on the sky compared to the consituent observations.

    iod_orbit_members : `FittedOrbitMembers`
        Dataframe of orbit members with the following columns:
            "orbit_id" : Orbit ID, a uuid [str],
            "obs_id" : Observation IDs [str], one ID per row.
            "residual_ra_arcsec" : Residual (observed - expected) equatorial J2000 Right Ascension in arcseconds [float]
            "residual_dec_arcsec" : Residual (observed - expected) equatorial J2000 Declination in arcseconds [float]
            "chi2" : Observation's chi2 [float]
            "gauss_sol" : Flag to indicate which observations were used to calculate the Gauss soluton [int]
            "outlier" : Flag to indicate which observations are potential outliers (their chi2 is higher than
                the chi2 threshold) [float]
    """
    # Initialize the propagator
    prop = propagator(**propagator_kwargs)

    processable = True
    if len(observations) == 0:
        processable = False

    obs_ids_all = observations.id.to_numpy(zero_copy_only=False)
    coords_all = observations.coordinates
    observers = observations.observers

    observations = observations.sort_by(
        [
            "coordinates.time.days",
            "coordinates.time.nanos",
            "coordinates.origin.code",
        ]
    )
    observers = observers.sort_by(
        ["coordinates.time.days", "coordinates.time.nanos", "coordinates.origin.code"]
    )

    coords_obs_all = observers.coordinates.r
    times_all = coords_all.time.mjd().to_numpy(zero_copy_only=False)

    orbit_sol: FittedOrbits = FittedOrbits.empty()
    obs_ids_sol = None
    arc_length = None
    outliers = np.array([])
    converged = False
    num_obs = len(observations)
    if num_obs < min_obs:
        processable = False

    max_outliers = calculate_max_outliers(num_obs, min_obs, contamination_percentage)

    # Select observation IDs to use for IOD
    obs_ids = select_observations(
        observations,
        method=observation_selection_method,
    )
    obs_ids = obs_ids[: (3 * (max_outliers + 1))]

    if len(obs_ids) == 0:
        processable = False

    j = 0
    while not converged and processable:
        if j == len(obs_ids):
            break

        ids = obs_ids[j]
        mask = np.isin(obs_ids_all, ids)

        # Grab sky-plane positions of the selected observations, the heliocentric ecliptic position of the observer,
        # and the times at which the observations occur
        coords = coords_all.apply_mask(mask)
        coords_obs = coords_obs_all[mask, :]
        times = times_all[mask]

        # Run IOD
        iod_orbits = gaussIOD(
            coords.values[:, 1:3],
            times,
            coords_obs,
            light_time=light_time,
            max_iter=100,
            tol=1e-15,
        )
        if len(iod_orbits) == 0:
            j += 1
            continue

        # Propagate initial orbit to all observation times
        ephemeris = prop.generate_ephemeris(
            iod_orbits, observers, chunk_size=1, max_processes=1
        )

        # For each unique initial orbit calculate residuals and chi-squared
        # Find the orbit which yields the lowest chi-squared
        orbit_ids = iod_orbits.orbit_id.to_numpy(zero_copy_only=False)
        for i, orbit_id in enumerate(orbit_ids):
            ephemeris_orbit = ephemeris.select("orbit_id", orbit_id)

            # Calculate residuals and chi2
            residuals = Residuals.calculate(
                coords_all,
                ephemeris_orbit.coordinates,
            )
            chi2 = residuals.chi2.to_numpy()
            chi2_total = np.sum(chi2)
            rchi2 = chi2_total / (2 * num_obs - 6)

            # The reduced chi2 is above the threshold and no outliers are
            # allowed, this cannot be improved by outlier rejection
            # so continue to the next IOD orbit
            if rchi2 > rchi2_threshold and max_outliers == 0:
                # If we have iterated through all iod orbits and no outliers
                # are allowed for this linkage then no other combination of
                # observations will make it acceptable, so exit here.
                if (i + 1) == len(iod_orbits):
                    processable = False
                    break

                continue

            # If the total reduced chi2 is less than the threshold accept the orbit
            elif rchi2 <= rchi2_threshold:
                logger.debug("Potential solution orbit has been found.")
                orbit_sol = iod_orbits[i : i + 1]
                obs_ids_sol = ids
                chi2_total_sol = chi2_total
                rchi2_sol = rchi2
                residuals_sol = residuals
                outliers = np.array([])
                arc_length = times_all.max() - times_all.min()
                converged = True
                break

            # Let's now test to see if we can remove some outliers, we
            # anticipate that we get to this stage if the three selected observations
            # belonging to one object yield a good initial orbit but the presence of outlier
            # observations is skewing the sum total of the residuals and chi2
            elif max_outliers > 0:
                logger.debug("Attempting to identify possible outliers.")
                for o in range(max_outliers):
                    # Select i highest observations that contribute to
                    # chi2 (and thereby the residuals)
                    remove = chi2[~mask].argsort()[-(o + 1) :]

                    # Grab the obs_ids for these outliers
                    obs_id_outlier = obs_ids_all[~mask][remove]
                    logger.debug("Possible outlier(s): {}".format(obs_id_outlier))

                    # Subtract the outlier's chi2 contribution
                    # from the total chi2
                    # Then recalculate the reduced chi2
                    chi2_new = chi2_total - np.sum(chi2[~mask][remove])
                    num_obs_new = len(observations) - len(remove)
                    rchi2_new = chi2_new / (2 * num_obs_new - 6)

                    ids_mask = np.isin(obs_ids_all, obs_id_outlier, invert=True)
                    arc_length = times_all[ids_mask].max() - times_all[ids_mask].min()

                    # If the updated reduced chi2 total is lower than our desired
                    # threshold, accept the soluton. If not, keep going.
                    if rchi2_new <= rchi2_threshold and arc_length >= min_arc_length:
                        orbit_sol = iod_orbits[i : i + 1]
                        obs_ids_sol = ids
                        chi2_total_sol = chi2_new
                        rchi2_sol = rchi2_new
                        residuals_sol = residuals
                        outliers = obs_id_outlier
                        num_obs = num_obs_new
                        ids_mask = np.isin(obs_ids_all, outliers, invert=True)
                        arc_length = (
                            times_all[ids_mask].max() - times_all[ids_mask].min()
                        )
                        converged = True
                        break

            else:
                continue

        j += 1

    if not converged or not processable:
        return FittedOrbits.empty(), FittedOrbitMembers.empty()

    else:
        orbit = FittedOrbits.from_kwargs(
            orbit_id=orbit_sol.orbit_id,
            object_id=orbit_sol.object_id,
            coordinates=orbit_sol.coordinates,
            arc_length=[arc_length],
            num_obs=[num_obs],
            chi2=[chi2_total_sol],
            reduced_chi2=[rchi2_sol],
        )

        orbit_members = FittedOrbitMembers.from_kwargs(
            orbit_id=np.full(
                len(obs_ids_all), orbit_sol.orbit_id[0].as_py(), dtype="object"
            ),
            obs_id=obs_ids_all,
            residuals=residuals_sol,
            solution=np.isin(obs_ids_all, obs_ids_sol),
            outlier=np.isin(obs_ids_all, outliers),
        )

    return orbit, orbit_members


def initial_orbit_determination(
    observations: Union[OrbitDeterminationObservations, ray.ObjectRef],
    linkage_members: Union[FittedOrbitMembers, ray.ObjectRef],
    propagator: Type[Propagator],
    min_obs: int = 6,
    min_arc_length: float = 1.0,
    contamination_percentage: float = 20.0,
    rchi2_threshold: float = 10**3,
    observation_selection_method: Literal[
        "combinations", "first+middle+last", "thirds"
    ] = "combinations",
    iterate: bool = False,
    light_time: bool = True,
    linkage_id_col: str = "cluster_id",
    propagator_kwargs: dict = {},
    chunk_size: int = 1,
    max_processes: Optional[int] = 1,
) -> Tuple[FittedOrbits, FittedOrbitMembers]:
    """
    Run initial orbit determination on linkages found in observations.
    """
    time_start = time.perf_counter()
    logger.info("Running initial orbit determination...")

    if isinstance(linkage_members, ray.ObjectRef):
        linkage_members_ref = linkage_members
        linkage_members = ray.get(linkage_members)
        logger.info("Retrieved linkage members from the object store.")
    else:
        linkage_members_ref = None

    if isinstance(observations, ray.ObjectRef):
        observations_ref = observations
        observations = ray.get(observations)
        logger.info("Retrieved observations from the object store.")
    else:
        observations_ref = None

    iod_orbits = FittedOrbits.empty()
    iod_orbit_members = FittedOrbitMembers.empty()
    if len(observations) > 0 and len(linkage_members) > 0:
        # Extract linkage IDs
        linkage_ids = linkage_members.column(linkage_id_col).unique()

        if max_processes is None:
            max_processes = mp.cpu_count()

        use_ray = initialize_use_ray(num_cpus=max_processes)
        if use_ray:
            refs_to_free = []

            linkage_ids_ref = ray.put(linkage_ids)
            refs_to_free.append(linkage_ids_ref)
            logger.info("Placed linkage IDs in the object store.")

            if linkage_members_ref is None:
                linkage_members_ref = ray.put(linkage_members)
                refs_to_free.append(linkage_members_ref)
                logger.info("Placed linkage members in the object store.")

            if observations_ref is None:
                observations_ref = ray.put(observations)
                refs_to_free.append(observations_ref)
                logger.info("Placed observations in the object store.")

            futures = []
            for linkage_id_chunk_indices in _iterate_chunk_indices(
                linkage_ids, chunk_size
            ):
                futures.append(
                    iod_worker_remote.remote(
                        linkage_ids_ref,
                        linkage_id_chunk_indices,
                        observations_ref,
                        linkage_members_ref,
                        min_obs=min_obs,
                        min_arc_length=min_arc_length,
                        contamination_percentage=contamination_percentage,
                        rchi2_threshold=rchi2_threshold,
                        observation_selection_method=observation_selection_method,
                        iterate=iterate,
                        light_time=light_time,
                        linkage_id_col=linkage_id_col,
                        propagator=propagator,
                        propagator_kwargs=propagator_kwargs,
                    )
                )

                if len(futures) >= max_processes * 1.5:
                    finished, futures = ray.wait(futures, num_returns=1)
                    result = ray.get(finished[0])
                    iod_orbits_chunk, iod_orbit_members_chunk = result
                    iod_orbits = qv.concatenate([iod_orbits, iod_orbits_chunk])
                    iod_orbit_members = qv.concatenate(
                        [iod_orbit_members, iod_orbit_members_chunk]
                    )
                    if iod_orbits.fragmented():
                        iod_orbits = qv.defragment(iod_orbits)
                    if iod_orbit_members.fragmented():
                        iod_orbit_members = qv.defragment(iod_orbit_members)

            while futures:
                finished, futures = ray.wait(futures, num_returns=1)
                result = ray.get(finished[0])
                iod_orbits_chunk, iod_orbit_members_chunk = result
                iod_orbits = qv.concatenate([iod_orbits, iod_orbits_chunk])
                iod_orbit_members = qv.concatenate(
                    [iod_orbit_members, iod_orbit_members_chunk]
                )
                if iod_orbits.fragmented():
                    iod_orbits = qv.defragment(iod_orbits)
                if iod_orbit_members.fragmented():
                    iod_orbit_members = qv.defragment(iod_orbit_members)

            if len(refs_to_free) > 0:
                ray.internal.free(refs_to_free)
                logger.info(
                    f"Removed {len(refs_to_free)} references from the object store."
                )

        else:
            for linkage_id_chunk in _iterate_chunks(linkage_ids, chunk_size):
                iod_orbits_chunk, iod_orbit_members_chunk = iod_worker(
                    linkage_id_chunk,
                    observations,
                    linkage_members,
                    min_obs=min_obs,
                    min_arc_length=min_arc_length,
                    contamination_percentage=contamination_percentage,
                    rchi2_threshold=rchi2_threshold,
                    observation_selection_method=observation_selection_method,
                    iterate=iterate,
                    light_time=light_time,
                    linkage_id_col=linkage_id_col,
                    propagator=propagator,
                    propagator_kwargs=propagator_kwargs,
                )
                iod_orbits = qv.concatenate([iod_orbits, iod_orbits_chunk])
                iod_orbit_members = qv.concatenate(
                    [iod_orbit_members, iod_orbit_members_chunk]
                )
                if iod_orbits.fragmented():
                    iod_orbits = qv.defragment(iod_orbits)
                if iod_orbit_members.fragmented():
                    iod_orbit_members = qv.defragment(iod_orbit_members)

        time_start_drop = time.time()
        logger.info("Removing duplicate initial orbits...")
        num_orbits = len(iod_orbits)
        iod_orbits, iod_orbit_members = drop_duplicate_orbits(
            iod_orbits,
            iod_orbit_members,
            subset=[
                "coordinates.time.days",
                "coordinates.time.nanos",
                "coordinates.x",
                "coordinates.y",
                "coordinates.z",
                "coordinates.vx",
                "coordinates.vy",
                "coordinates.vz",
            ],
            keep="first",
        )
        time_end_drop = time.time()
        logger.info(f"Removed {num_orbits - len(iod_orbits)} duplicate clusters.")
        time_end_drop = time.time()
        logger.info(
            f"Inital orbit deduplication completed in {time_end_drop - time_start_drop:.3f} seconds."
        )

        # Sort initial orbits by orbit ID and observation time
        iod_orbits, iod_orbit_members = sort_by_id_and_time(
            iod_orbits, iod_orbit_members, observations, "orbit_id"
        )

    else:
        iod_orbits = FittedOrbits.empty()
        iod_orbit_members = FittedOrbitMembers.empty()

    time_end = time.perf_counter()
    logger.info(f"Found {len(iod_orbits)} initial orbits.")
    logger.info(
        f"Initial orbit determination completed in {time_end - time_start:.3f} seconds."
    )

    return iod_orbits, iod_orbit_members



import numpy as np

from adam_core.constants import Constants as c

__all__ = ["calcGibbs"]

MU = c.MU


def calcGibbs(r1, r2, r3):
    """
    Calculates the velocity vector at the location of the second position vector (r2) using the
    Gibbs method.

    .. math::
        \vec{D} = \vec{r}_1 \times \vec{r}_2  +  \vec{r}_2 \times \vec{r}_3 +  \vec{r}_3 \times \vec{r}_1

        \vec{N} = r_1 (\vec{r}_2 \times \vec{r}_3) + r_2 (\vec{r}_3 \times \vec{r}_1) + r_3 (\vec{r}_1 \times \vec{r}_2)

        \vec{B} \equiv \vec{D} \times \vec{r}_2

        L_g \equiv \sqrt{\frac{\mu}{ND}}

        \vec{v}_2 = \frac{L_g}{r_2} \vec{B} + L_g \vec{S}

    For more details on theory see Chapter 4 in David A. Vallado's "Fundamentals of Astrodynamics
    and Applications".

    Parameters
    ----------
    r1 : `~numpy.ndarray` (3)
        Heliocentric position vector at time 1 in cartesian coordinates in units
        of AU.
    r2 : `~numpy.ndarray` (3)
        Heliocentric position vector at time 2 in cartesian coordinates in units
        of AU.
    r3 : `~numpy.ndarray` (3)
        Heliocentric position vector at time 3 in cartesian coordinates in units
        of AU.

    Returns
    -------
    v2 : `~numpy.ndarray` (3)
        Velocity of object at position r2 at time t2 in units of AU per day.
    """
    r1_mag = np.linalg.norm(r1)
    r2_mag = np.linalg.norm(r2)
    r3_mag = np.linalg.norm(r3)
    Z12 = np.cross(r1, r2)
    Z23 = np.cross(r2, r3)
    Z31 = np.cross(r3, r1)

    # coplanarity = np.arcsin(np.dot(Z23, r1) / (np.linalg.norm(Z23) * r1_mag))

    N = r1_mag * Z23 + r2_mag * Z31 + r3_mag * Z12
    N_mag = np.linalg.norm(N)
    D = Z12 + Z23 + Z31
    D_mag = np.linalg.norm(D)
    S = (r2_mag - r3_mag) * r1 + (r3_mag - r1_mag) * r2 + (r1_mag - r2_mag) * r3
    # S_mag = np.linalg.norm(S)
    B = np.cross(D, r2)
    Lg = np.sqrt(MU / N_mag / D_mag)
    v2 = Lg / r2_mag * B + Lg * S
    return v2



import numpy as np
import pytest

from adam_core.coordinates.cartesian import CartesianCoordinates
from adam_core.coordinates.origin import Origin, OriginCodes
from adam_core.observers.observers import Observers
from adam_core.photometry import calculate_phase_angle
from adam_core.time import Timestamp


def _make_helio_cartesian(
    r_au: np.ndarray,
    *,
    origin: OriginCodes = OriginCodes.SUN,
    frame: str = "ecliptic",
) -> CartesianCoordinates:
    """
    Build a `CartesianCoordinates` table with zero velocities.

    Why: `calculate_phase_angle` only needs positions, but the coordinate type is a 6D state.
    """
    r = np.asarray(r_au, dtype=np.float64)
    if r.ndim != 2 or r.shape[1] != 3:
        raise ValueError("r_au must have shape (N, 3)")

    n = int(r.shape[0])
    t = Timestamp.from_mjd([60000.0] * n, scale="tdb")
    origin_col = Origin.from_kwargs(code=[origin.name] * n)

    zeros = np.zeros(n, dtype=np.float64)
    return CartesianCoordinates.from_kwargs(
        x=r[:, 0],
        y=r[:, 1],
        z=r[:, 2],
        vx=zeros,
        vy=zeros,
        vz=zeros,
        time=t,
        origin=origin_col,
        frame=frame,
    )


def _phase_angle_reference_deg(
    object_pos_au: np.ndarray, observer_pos_au: np.ndarray
) -> np.ndarray:
    """
    Reference phase-angle computation via vector angle definition.

    Phase angle is the angle at the object between:
    - Sun direction (object -> Sun), and
    - Observer direction (object -> observer).

    We compute the angle between vectors using atan2(||a×b||, a·b), which is stable near 0/180.
    """
    obj = np.asarray(object_pos_au, dtype=np.float64)
    obs = np.asarray(observer_pos_au, dtype=np.float64)
    a = -obj
    b = obs - obj
    cross = np.cross(a, b)
    dot = np.sum(a * b, axis=1)
    alpha_rad = np.arctan2(np.linalg.norm(cross, axis=1), dot)
    return np.degrees(alpha_rad)


def test_calculate_phase_angle_expected_values() -> None:
    # Three simple heliocentric geometries with known phase angles:
    # - 0 deg (opposition-like): Sun -> observer -> object
    # - 90 deg (quadrature-like)
    # - 180 deg (conjunction-like): Sun -> object -> observer
    t = Timestamp.from_mjd([60000.0, 60000.0, 60000.0], scale="tdb")
    origin_sun = Origin.from_kwargs(code=[OriginCodes.SUN.name] * 3)

    object_coords = CartesianCoordinates.from_kwargs(
        x=[2.0, 1.0, 1.0],
        y=[0.0, 0.0, 0.0],
        z=[0.0, 0.0, 0.0],
        vx=[0.0, 0.0, 0.0],
        vy=[0.0, 0.0, 0.0],
        vz=[0.0, 0.0, 0.0],
        time=t,
        origin=origin_sun,
        frame="ecliptic",
    )
    observer_coords = CartesianCoordinates.from_kwargs(
        x=[1.0, 1.0, 2.0],
        y=[0.0, 1.0, 0.0],
        z=[0.0, 0.0, 0.0],
        vx=[0.0, 0.0, 0.0],
        vy=[0.0, 0.0, 0.0],
        vz=[0.0, 0.0, 0.0],
        time=t,
        origin=origin_sun,
        frame="ecliptic",
    )
    observers = Observers.from_kwargs(code=["X"] * 3, coordinates=observer_coords)

    alpha = calculate_phase_angle(object_coords, observers)
    np.testing.assert_allclose(alpha, np.array([0.0, 90.0, 180.0]), atol=1e-12)


def test_calculate_phase_angle_length_mismatch_raises() -> None:
    t_obj = Timestamp.from_mjd([60000.0, 60000.0], scale="tdb")
    t_obs = Timestamp.from_mjd([60000.0], scale="tdb")

    origin_obj = Origin.from_kwargs(code=[OriginCodes.SUN.name] * 2)
    origin_obs = Origin.from_kwargs(code=[OriginCodes.SUN.name])

    object_coords = CartesianCoordinates.from_kwargs(
        x=[1.0, 1.0],
        y=[0.0, 0.0],
        z=[0.0, 0.0],
        vx=[0.0, 0.0],
        vy=[0.0, 0.0],
        vz=[0.0, 0.0],
        time=t_obj,
        origin=origin_obj,
        frame="ecliptic",
    )
    observer_coords = CartesianCoordinates.from_kwargs(
        x=[1.0],
        y=[0.0],
        z=[0.0],
        vx=[0.0],
        vy=[0.0],
        vz=[0.0],
        time=t_obs,
        origin=origin_obs,
        frame="ecliptic",
    )
    observers = Observers.from_kwargs(code=["X"], coordinates=observer_coords)

    with pytest.raises(ValueError, match="must match object_coords length"):
        calculate_phase_angle(object_coords, observers)


def test_calculate_phase_angle_requires_heliocentric_origin() -> None:
    t = Timestamp.from_mjd([60000.0], scale="tdb")

    obj_ssb = CartesianCoordinates.from_kwargs(
        x=[1.0],
        y=[0.0],
        z=[0.0],
        vx=[0.0],
        vy=[0.0],
        vz=[0.0],
        time=t,
        origin=Origin.from_kwargs(code=[OriginCodes.SOLAR_SYSTEM_BARYCENTER.name]),
        frame="ecliptic",
    )
    obs_sun = CartesianCoordinates.from_kwargs(
        x=[1.0],
        y=[0.0],
        z=[0.0],
        vx=[0.0],
        vy=[0.0],
        vz=[0.0],
        time=t,
        origin=Origin.from_kwargs(code=[OriginCodes.SUN.name]),
        frame="ecliptic",
    )
    observers = Observers.from_kwargs(code=["X"], coordinates=obs_sun)

    with pytest.raises(ValueError, match="object_coords must be heliocentric"):
        calculate_phase_angle(obj_ssb, observers)


def test_calculate_phase_angle_requires_heliocentric_observer_origin() -> None:
    t = Timestamp.from_mjd([60000.0], scale="tdb")

    obj_sun = CartesianCoordinates.from_kwargs(
        x=[1.0],
        y=[0.0],
        z=[0.0],
        vx=[0.0],
        vy=[0.0],
        vz=[0.0],
        time=t,
        origin=Origin.from_kwargs(code=[OriginCodes.SUN.name]),
        frame="ecliptic",
    )
    obs_ssb = CartesianCoordinates.from_kwargs(
        x=[1.0],
        y=[0.0],
        z=[0.0],
        vx=[0.0],
        vy=[0.0],
        vz=[0.0],
        time=t,
        origin=Origin.from_kwargs(code=[OriginCodes.SOLAR_SYSTEM_BARYCENTER.name]),
        frame="ecliptic",
    )
    observers = Observers.from_kwargs(code=["X"], coordinates=obs_ssb)

    with pytest.raises(
        ValueError, match="observers\\.coordinates must be heliocentric"
    ):
        calculate_phase_angle(obj_sun, observers)


def test_calculate_phase_angle_requires_same_frame() -> None:
    t = Timestamp.from_mjd([60000.0], scale="tdb")
    origin_sun = Origin.from_kwargs(code=[OriginCodes.SUN.name])

    obj = CartesianCoordinates.from_kwargs(
        x=[1.0],
        y=[0.0],
        z=[0.0],
        vx=[0.0],
        vy=[0.0],
        vz=[0.0],
        time=t,
        origin=origin_sun,
        frame="ecliptic",
    )
    obs = CartesianCoordinates.from_kwargs(
        x=[1.0],
        y=[0.0],
        z=[0.0],
        vx=[0.0],
        vy=[0.0],
        vz=[0.0],
        time=t,
        origin=origin_sun,
        frame="equatorial",
    )
    observers = Observers.from_kwargs(code=["X"], coordinates=obs)

    with pytest.raises(ValueError, match="same frame"):
        calculate_phase_angle(obj, observers)


@pytest.mark.parametrize(
    "object_pos, observer_pos",
    [
        # r == 0 (object at Sun)
        (np.array([[0.0, 0.0, 0.0]]), np.array([[1.0, 0.0, 0.0]])),
        # delta == 0 (observer colocated with object)
        (np.array([[1.0, 0.0, 0.0]]), np.array([[1.0, 0.0, 0.0]])),
        # non-finite values
        (np.array([[np.nan, 0.0, 0.0]]), np.array([[1.0, 0.0, 0.0]])),
    ],
)
def test_calculate_phase_angle_rejects_invalid_geometry(
    object_pos: np.ndarray, observer_pos: np.ndarray
) -> None:
    obj = _make_helio_cartesian(object_pos, origin=OriginCodes.SUN, frame="ecliptic")
    obs = _make_helio_cartesian(observer_pos, origin=OriginCodes.SUN, frame="ecliptic")
    observers = Observers.from_kwargs(code=["X"], coordinates=obs)

    with pytest.raises(ValueError, match="Invalid photometry geometry"):
        calculate_phase_angle(obj, observers)


def test_calculate_phase_angle_matches_vector_definition_for_random_cases() -> None:
    rng = np.random.default_rng(0)

    # Generate a deterministic set of non-degenerate geometries.
    obj_list: list[np.ndarray] = []
    obs_list: list[np.ndarray] = []
    while len(obj_list) < 200:
        obj = rng.uniform(-3.0, 3.0, size=3)
        obs = rng.uniform(-3.0, 3.0, size=3)
        if np.linalg.norm(obj) <= 0.2:
            continue
        if np.linalg.norm(obs) <= 0.2:
            continue
        if np.linalg.norm(obj - obs) <= 0.2:
            continue
        obj_list.append(obj)
        obs_list.append(obs)

    object_pos = np.stack(obj_list, axis=0)
    observer_pos = np.stack(obs_list, axis=0)

    obj_coords = _make_helio_cartesian(
        object_pos, origin=OriginCodes.SUN, frame="ecliptic"
    )
    obs_coords = _make_helio_cartesian(
        observer_pos, origin=OriginCodes.SUN, frame="ecliptic"
    )
    observers = Observers.from_kwargs(
        code=["X"] * len(obj_coords), coordinates=obs_coords
    )

    alpha = calculate_phase_angle(obj_coords, observers)
    alpha_ref = _phase_angle_reference_deg(object_pos, observer_pos)

    assert alpha.dtype == np.float64
    assert alpha.shape == (len(obj_coords),)
    assert np.all(np.isfinite(alpha))
    assert np.all(alpha >= 0.0)
    assert np.all(alpha <= 180.0)

    np.testing.assert_allclose(alpha, alpha_ref, rtol=0.0, atol=1e-12)



from __future__ import annotations

import os
from pathlib import Path

import numpy as np
import pyarrow as pa
import pyarrow.compute as pc
import pytest

from adam_core.coordinates.cartesian import CartesianCoordinates
from adam_core.coordinates.origin import Origin
from adam_core.observations.detections import PointSourceDetections
from adam_core.observations.exposures import Exposures
from adam_core.observers.observers import Observers
from adam_core.photometry.absolute_magnitude import (
    estimate_absolute_magnitude_v_from_detections,
)
from adam_core.photometry.bandpasses.api import map_to_canonical_filter_bands
from adam_core.photometry.magnitude import predict_magnitudes
from adam_core.time import Timestamp

DATA_DIR = Path(__file__).parent / "data"

MPC_FIXTURES: list[str] = sorted(
    p.name for p in DATA_DIR.glob("mpc_magnitude_fixture_*.npz")
)
if not MPC_FIXTURES:
    MPC_FIXTURES = ["__NO_FIXTURES__"]


def _observers_from_heliocentric_positions(
    station_code: str, times: Timestamp, heliocentric_pos_au: np.ndarray
) -> Observers:
    pos = np.asarray(heliocentric_pos_au, dtype=np.float64)
    if pos.ndim != 2 or pos.shape[1] != 3:
        raise ValueError("heliocentric_pos_au must have shape (N, 3)")
    n = int(pos.shape[0])
    return Observers.from_kwargs(
        code=[station_code] * n,
        coordinates=CartesianCoordinates.from_kwargs(
            x=pos[:, 0],
            y=pos[:, 1],
            z=pos[:, 2],
            vx=np.zeros(n),
            vy=np.zeros(n),
            vz=np.zeros(n),
            time=times,
            frame="ecliptic",
            origin=Origin.from_kwargs(code=["SUN"] * n),
        ),
    )


def _canonicalize_exposure_filters(
    exposures: Exposures, *, strict: bool = False
) -> Exposures:
    canonical = map_to_canonical_filter_bands(
        exposures.observatory_code,
        exposures.filter,
        allow_fallback_filters=not strict,
    )
    return exposures.set_column(
        "filter", pa.array(canonical.tolist(), type=pa.large_string())
    )


def _sse(resid: np.ndarray) -> float:
    x = np.asarray(resid, dtype=np.float64)
    return float(np.sum(x * x))


def _phase_angle_deg(object_pos: np.ndarray, observer_pos: np.ndarray) -> np.ndarray:
    """
    Phase angle (Sun-object-observer) in degrees, matching the H-G geometry used by the model.

    Inputs are heliocentric vectors in AU with shape (N, 3).
    """
    obj = np.asarray(object_pos, dtype=np.float64)
    obs = np.asarray(observer_pos, dtype=np.float64)
    if obj.ndim != 2 or obs.ndim != 2 or obj.shape != obs.shape or obj.shape[1] != 3:
        raise ValueError("object_pos and observer_pos must both have shape (N, 3)")

    r = np.sqrt(np.sum(obj * obj, axis=1))
    delta_vec = obj - obs
    delta = np.sqrt(np.sum(delta_vec * delta_vec, axis=1))
    obs_sun = np.sqrt(np.sum(obs * obs, axis=1))
    numer = r**2 + delta**2 - obs_sun**2
    denom = 2.0 * r * delta
    cos_phase = np.clip(numer / denom, -1.0, 1.0)
    return np.degrees(np.arccos(cos_phase))


def _format_residual_table(rows: list[dict[str, object]], cols: list[str]) -> str:
    widths: dict[str, int] = {c: len(c) for c in cols}
    for r in rows:
        for c in cols:
            widths[c] = max(widths[c], len(str(r.get(c, ""))))

    def fmt_row(r: dict[str, object]) -> str:
        parts = [str(r.get(c, "")).rjust(widths[c]) for c in cols]
        return "  ".join(parts)

    header = "  ".join([c.rjust(widths[c]) for c in cols])
    sep = "  ".join(["-" * widths[c] for c in cols])
    body = "\n".join(fmt_row(r) for r in rows)
    return "\n".join([header, sep, body])


@pytest.mark.parametrize("fixture_name", MPC_FIXTURES)
def test_mpc_absolute_magnitude_from_fixture(
    monkeypatch, fixture_name: str, pytestconfig, capsys: pytest.CaptureFixture[str]
) -> None:
    if fixture_name == "__NO_FIXTURES__":
        pytest.skip("No photometry fixtures found on disk.")

    fixture_path = DATA_DIR / fixture_name
    if not fixture_path.exists():
        pytest.skip(f"Missing fixture {fixture_name}")

    fx = np.load(fixture_path, allow_pickle=True)
    station = str(fx["station"][0])
    H_v_mpc = float(fx["H_v_mpc"][0])
    G_mpc = float(fx["G_mpc"][0])
    time_iso = fx["time_iso"].astype(object).tolist()
    bands = fx["band"].astype(object).tolist()
    mag_obs = np.asarray(fx["mag_obs"], dtype=np.float64)
    obj_pos = np.asarray(fx["object_pos"], dtype=np.float64)
    obs_pos = np.asarray(fx["observer_pos"], dtype=np.float64)

    n = len(mag_obs)
    assert obj_pos.shape == (n, 3)
    assert obs_pos.shape == (n, 3)
    assert len(bands) == n

    times = Timestamp.from_iso8601(time_iso, scale="utc")

    # Build a SMALL exposures table (unique by band) so the estimator must join via exposure_id
    # and expand back to per-detection rows internally.
    uniq_bands = sorted({str(b).strip() for b in bands})
    exposure_ids = [f"e_{b}" for b in uniq_bands]
    exposures = Exposures.from_kwargs(
        id=exposure_ids,
        start_time=Timestamp.from_iso8601(
            [time_iso[0]] * len(exposure_ids), scale="utc"
        ),
        duration=np.zeros(len(exposure_ids), dtype=np.float64),
        filter=uniq_bands,
        observatory_code=[station] * len(exposure_ids),
    )

    # Critical sanity: patch the *class* method so it applies to Exposures objects created
    # inside the estimator via .take()/.set_column(). Use fixture-provided observer_pos.
    def fake_observers(self, *args, **kwargs):  # noqa: ARG001
        if len(self) != n:
            raise AssertionError(f"Expected exposures length {n}, got {len(self)}")
        return _observers_from_heliocentric_positions(station, times, obs_pos)

    monkeypatch.setattr(Exposures, "observers", fake_observers)

    object_coords = CartesianCoordinates.from_kwargs(
        x=obj_pos[:, 0],
        y=obj_pos[:, 1],
        z=obj_pos[:, 2],
        vx=np.zeros(n),
        vy=np.zeros(n),
        vz=np.zeros(n),
        time=times,
        frame="ecliptic",
        origin=Origin.from_kwargs(code=["SUN"] * n),
    )

    detections = PointSourceDetections.from_kwargs(
        id=[f"d{i}" for i in range(n)],
        exposure_id=[f"e_{str(b).strip()}" for b in bands],
        time=times,
        ra=np.zeros(n),
        dec=np.zeros(n),
        mag=mag_obs,
        mag_sigma=None,
    )

    pp_hat = estimate_absolute_magnitude_v_from_detections(
        detections,
        exposures,
        object_coords,
        composition="NEO",
        G=G_mpc,
        strict_band_mapping=False,
        reference_filter="V",
    )
    H_hat = float(pp_hat.H_v[0].as_py())
    assert np.isfinite(H_hat)

    # For evaluation, align + canonicalize exposures to get a valid forward-model run.
    idx = pc.fill_null(pc.index_in(detections.exposure_id, value_set=exposures.id), -1)
    idx_np = np.asarray(idx.to_numpy(zero_copy_only=False), dtype=np.int32)
    if np.any(idx_np < 0):
        raise AssertionError("fixture test exposure_id mapping failed unexpectedly")
    exposures_aligned = exposures.take(pa.array(idx_np, type=pa.int32()))
    exposures_canon = _canonicalize_exposure_filters(exposures_aligned, strict=False)

    # Compare fit quality using a single forward-model call at H=0.
    m0 = predict_magnitudes(
        H=0.0,
        object_coords=object_coords,
        exposures=exposures_canon,
        G=G_mpc,
        reference_filter="V",
        composition="NEO",
    )
    resid_hat = (np.asarray(m0, dtype=np.float64) + H_hat) - mag_obs
    resid_mpc = (np.asarray(m0, dtype=np.float64) + H_v_mpc) - mag_obs

    # The estimator is least-squares optimal in H for fixed geometry, G, and composition.
    assert _sse(resid_hat) <= _sse(resid_mpc) + 1e-10

    # Strong sanity check: with missing per-point sigma, the optimal H is mean(mag - m0).
    H_opt = float(np.mean(mag_obs - np.asarray(m0, dtype=np.float64)))
    assert H_hat == pytest.approx(H_opt, abs=1e-12)

    verbose = bool(pytestconfig.getoption("--photometry-fixtures-verbose")) or bool(
        os.environ.get("ADAM_CORE_PHOTOMETRY_FIXTURE_VERBOSE")
    )
    if verbose:
        canon = np.asarray(
            exposures_canon.filter.to_numpy(zero_copy_only=False), dtype=object
        ).astype(str)
        phase_deg = _phase_angle_deg(obj_pos, obs_pos)

        cols = [
            "H_mpc",
            "H_hat",
            "dH",
            "filter_id",
            "n",
            "phase_deg_span",
            "abs_resid_median_hat",
            "abs_resid_p95_hat",
            "resid_mean_hat",
            "abs_resid_median_mpc",
            "abs_resid_p95_mpc",
            "resid_mean_mpc",
        ]
        rows: list[dict[str, object]] = []
        for k in sorted(set(canon.tolist())):
            m = canon == str(k)
            xs_hat = np.asarray(resid_hat[m], dtype=np.float64)
            xs_mpc = np.asarray(resid_mpc[m], dtype=np.float64)
            ph = np.asarray(phase_deg[m], dtype=np.float64)
            rows.append(
                {
                    "H_mpc": f"{H_v_mpc:.4f}",
                    "H_hat": f"{H_hat:.4f}",
                    "dH": f"{(H_hat - H_v_mpc):+.4f}",
                    "filter_id": str(k),
                    "n": int(xs_hat.size),
                    "phase_deg_span": f"{float(np.ptp(ph)):.2f}" if ph.size else "nan",
                    "abs_resid_median_hat": f"{float(np.median(np.abs(xs_hat))):.6f}",
                    "abs_resid_p95_hat": f"{float(np.quantile(np.abs(xs_hat), 0.95)):.6f}",
                    "resid_mean_hat": f"{float(np.mean(xs_hat)):.6f}",
                    "abs_resid_median_mpc": f"{float(np.median(np.abs(xs_mpc))):.6f}",
                    "abs_resid_p95_mpc": f"{float(np.quantile(np.abs(xs_mpc), 0.95)):.6f}",
                    "resid_mean_mpc": f"{float(np.mean(xs_mpc)):.6f}",
                }
            )
        rows.sort(key=lambda r: str(r["filter_id"]))
        table = _format_residual_table(rows, cols)
        with capsys.disabled():
            print(
                "\n".join(
                    [
                        f"\nAbs-mag fixture summary for {fixture_name} (station={station}):",
                        f"H_mpc={H_v_mpc:.4f}, G_mpc={G_mpc:.4f}, H_hat={H_hat:.4f}, dH={H_hat - H_v_mpc:+.4f}",
                        table,
                        "",
                    ]
                )
            )


"""
NOTE: X05/LSST is covered by the unified MPC-format fixtures (station='X05')
generated via `generate_mpc_magnitude_fixtures.py`; we do not maintain a separate
LSST-only fixture schema/test here.
"""



from typing import Optional

import numpy as np

from ..coordinates import OriginCodes
from ..orbits import Orbits
from ..propagator import Propagator
from ..time import Timestamp
from ..utils.plots.logos import AsteroidInstituteLogoDark, get_logo_base64
from ..utils.spice import get_perturber_state

try:
    import plotly.graph_objects as go
except ImportError:
    raise ImportError("Please install adam_core[plots] to use this feature.")


def plot_orbit(
    orbit: Orbits,
    propagator: Propagator,
    start_time: Optional[Timestamp] = None,
    logo: bool = True,
) -> go.Figure:
    """
    Plot an orbit.

    Parameters
    ----------
    orbit: Orbits
        The orbit to plot.
    propagator: Propagator
        The propagator to use to propagate the orbit.
    start_time: Optional[Timestamp]
        The start time to use for the propagation.
    logo: bool, optional
        Whether to add the Asteroid Institute logo to the plot.
    """
    # Get the period so we know how long to propagate in cartesian space
    keplerian = orbit.coordinates.to_keplerian()

    period_index = int(np.argmax(keplerian.P))
    period = keplerian.P[period_index]

    if start_time is None:
        start_time = orbit.coordinates.time[period_index]

    # Add the period in days to the orbit.coordinates.time
    propagation_end_date = (
        start_time.add_days(np.ceil(period).astype(int)).mjd()[0].as_py()
    )

    sample_dates_mjd = np.arange(
        start_time.mjd()[0].as_py(), propagation_end_date + 5, 5
    )
    sample_dates = Timestamp.from_mjd(sample_dates_mjd, scale=start_time.scale)

    # Propagate the orbit in cartesian space
    propagated_orbits = propagator.propagate_orbits(orbit, sample_dates)

    max_r = np.max(np.abs(propagated_orbits.coordinates.r))

    # Give a default max_r if the orbits are smaller than the inner planets
    max_r = np.max((max_r, 4))

    # Render the planets conditionally based on max r of the propagated orbits
    conditional_planet_distances = (
        ("Jupiter", 5, OriginCodes.JUPITER_BARYCENTER),
        ("Saturn", 10, OriginCodes.SATURN_BARYCENTER),
        ("Uranus", 15, OriginCodes.URANUS_BARYCENTER),
        ("Neptune", 20, OriginCodes.NEPTUNE_BARYCENTER),
    )

    planet_states = {
        "Earth": get_perturber_state(OriginCodes.EARTH, sample_dates),
        "Mars": get_perturber_state(OriginCodes.MARS_BARYCENTER, sample_dates),
        "Venus": get_perturber_state(OriginCodes.VENUS, sample_dates),
        "Mercury": get_perturber_state(OriginCodes.MERCURY, sample_dates),
    }

    for planet, distance, origin_code in conditional_planet_distances:
        if distance < max_r:
            planet_states[planet] = get_perturber_state(origin_code, sample_dates)

    traces = []

    # Create figure with black background and appropriate styling
    fig = go.Figure(
        layout=dict(
            paper_bgcolor="black",
            plot_bgcolor="black",
            scene=dict(
                bgcolor="black",
            ),
        )
    )

    # Define a color scheme for planets with transparency
    planet_colors = {
        "Mercury": "rgba(160, 82, 45, 0.6)",  # Brown with alpha
        "Venus": "rgba(218, 165, 32, 0.6)",  # Goldenrod with alpha
        "Earth": "rgba(65, 105, 225, 0.6)",  # Royal Blue with alpha
        "Mars": "rgba(205, 92, 92, 0.6)",  # Indian Red with alpha
        "Jupiter": "rgba(222, 184, 135, 0.6)",  # Burlywood with alpha
        "Saturn": "rgba(244, 164, 96, 0.6)",  # Sandy Brown with alpha
        "Uranus": "rgba(135, 206, 235, 0.6)",  # Sky Blue with alpha
        "Neptune": "rgba(30, 144, 255, 0.6)",  # Dodger Blue with alpha
    }

    for planet, state in planet_states.items():
        traces.append(
            go.Scatter3d(
                x=state.x,
                y=state.y,
                z=state.z,
                mode="lines",
                name=planet,
                line=dict(width=1, color=planet_colors[planet]),
            )
        )

    # Use bright green for all non-planet orbits
    orbit_color = "#00FF00"  # Bright green

    for orbit_id in orbit.orbit_id.unique():
        propagated_orbit = propagated_orbits.select("orbit_id", orbit_id)
        cartesian = propagated_orbit.coordinates
        isot_time = propagated_orbit.coordinates.time.to_astropy().isot
        traces.append(
            go.Scatter3d(
                x=cartesian.x,
                y=cartesian.y,
                z=cartesian.z,
                mode="lines",
                name=f"{propagated_orbit.orbit_id[0].as_py()} {propagated_orbit.object_id[0].as_py()}",
                hovertext=isot_time,
                line=dict(width=2, color=orbit_color),
            )
        )

    # For now we are always heliocentric, so plot the sun as a sphere
    traces.append(
        go.Scatter3d(
            x=[0],
            y=[0],
            z=[0],
            mode="markers",
            name="Sun",
            marker=dict(size=1, color="yellow"),
        )
    )

    for trace in traces:
        fig.add_trace(trace)

    max_r_padded = max_r * 1.1

    # Add logo if requested
    if logo:
        images = [
            dict(
                source=get_logo_base64(AsteroidInstituteLogoDark),
                xref="paper",
                yref="paper",
                x=1.02,
                y=-0.15,
                sizex=0.20,
                sizey=0.20,
                xanchor="left",
                yanchor="bottom",
                layer="above",
            )
        ]
    else:
        images = []

    fig.update(
        layout=dict(
            scene=dict(
                xaxis=dict(
                    range=[-max_r_padded, max_r_padded],
                    gridcolor="rgba(128, 128, 128, 0.2)",
                    showbackground=False,
                    color="white",
                ),
                yaxis=dict(
                    range=[-max_r_padded, max_r_padded],
                    gridcolor="rgba(128, 128, 128, 0.2)",
                    showbackground=False,
                    color="white",
                ),
                zaxis=dict(
                    range=[-max_r_padded, max_r_padded],
                    gridcolor="rgba(128, 128, 128, 0.2)",
                    showbackground=False,
                    color="white",
                ),
                xaxis_title="x [au]",
                yaxis_title="y [au]",
                zaxis_title="z [au]",
            ),
            font=dict(color="white"),  # Make all text white
            images=images,  # Add the logo images
        )
    )

    return fig



import logging
from enum import Enum

import numpy as np
import pyarrow as pa
import pyarrow.compute as pc
import quivr as qv

from ..constants import KM_P_AU, S_P_DAY

logger = logging.getLogger(__name__)


class OriginCodes(Enum):
    SOLAR_SYSTEM_BARYCENTER = 0
    MERCURY_BARYCENTER = 1
    VENUS_BARYCENTER = 2
    EARTH_MOON_BARYCENTER = 3
    MARS_BARYCENTER = 4
    JUPITER_BARYCENTER = 5
    SATURN_BARYCENTER = 6
    URANUS_BARYCENTER = 7
    NEPTUNE_BARYCENTER = 8
    SUN = 10
    MERCURY = 199
    VENUS = 299
    EARTH = 399
    MOON = 301
    MARS = 499
    JUPITER = 599
    SATURN = 699
    URANUS = 799
    NEPTUNE = 899


def _convert_mu_units(mu: float):
    """
    Convert mu in km^3 / s^2 to au^3 / day^2.
    """
    return mu / KM_P_AU**3 * S_P_DAY**2


class OriginGravitationalParameters(float, Enum):
    # Taken from https://iopscience.iop.org/article/10.3847/1538-3881/abd414
    MERCURY_BARYCENTER = _convert_mu_units(22032.080486418)
    VENUS_BARYCENTER = _convert_mu_units(324858.592000)
    MARS_BARYCENTER = _convert_mu_units(42828.375816)
    JUPITER_BARYCENTER = _convert_mu_units(126712764.100000)
    SATURN_BARYCENTER = _convert_mu_units(37940584.841800)
    URANUS_BARYCENTER = _convert_mu_units(5794556.400000)
    NEPTUNE_BARYCENTER = _convert_mu_units(6836527.100580)
    PLUTO_BARYCENTER = _convert_mu_units(975.500000)
    SUN = _convert_mu_units(132712440041.279419)
    MERCURY = _convert_mu_units(22031.868551)
    VENUS = _convert_mu_units(324858.592000)
    EARTH = _convert_mu_units(398600.435507)
    MOON = _convert_mu_units(4902.800118)

    @classmethod
    def SOLAR_SYSTEM_BARYCENTER(cls) -> float:
        """
        Return the effective gravitational parameter of the Solar System barycenter as approximated
        by adding the gravitational parameters of the Sun, Mercury, Venus, Earth, Moon,
        Mars, Jupiter, Uranus, and Neptune.

        This should only be used for calculating osculating elements with respect to the Solar System barycenter.

        Horizons's effective gravitational parameter (as of 2025-11-26) for osculating elements with respect to the Solar System barycenter
        is 1.3289051882019876E+11 km^3/s^2 which agrees to within 1e-6 with the value calculated here. Discrepancies are likely due to the inclusion
        of small bodies not included in the calculation here.

        Returns
        -------
        mu : float
            The gravitational parameter of the Solar System barycenter in au^3 / day^2.
        """
        return (
            cls.SUN
            + cls.MERCURY_BARYCENTER
            + cls.VENUS_BARYCENTER
            + cls.EARTH
            + cls.MOON
            + cls.MARS_BARYCENTER
            + cls.JUPITER_BARYCENTER
            + cls.SATURN_BARYCENTER
            + cls.URANUS_BARYCENTER
            + cls.NEPTUNE_BARYCENTER
            + cls.PLUTO_BARYCENTER
        )


# TODO: Replace with DictionaryColumn or similar
#       Investigate whether this class is even necessary
class Origin(qv.Table):
    code = qv.LargeStringColumn()

    @classmethod
    def from_OriginCodes(cls, code: OriginCodes, size: int = 1) -> "Origin":
        return cls.from_kwargs(
            code=pa.array(pa.repeat(code.name, size), type=pa.large_string())
        )

    def as_OriginCodes(self) -> OriginCodes:
        """
        Convert the origin codes to an `~adam_core.coordinates.origin.OriginCodes` object.

        Returns
        -------
        OriginCodes
            Origin codes as an `~adam_core.coordinates.origin.OriginCodes` object.
        """
        assert (
            len(self.code.unique()) == 1
        ), "Only one origin code can be converted at a time."
        return OriginCodes[self.code.unique()[0].as_py()]

    def __eq__(self, other: object) -> np.ndarray:
        if isinstance(other, (str, np.ndarray)):
            codes = self.code.to_numpy(zero_copy_only=False)
            return codes == other
        elif isinstance(other, OriginCodes):
            codes = self.code.to_numpy(zero_copy_only=False)
            return codes == other.name
        elif isinstance(other, Origin):
            codes = self.code.to_numpy(zero_copy_only=False)
            other_codes = other.code.to_numpy(zero_copy_only=False)
            return codes == other_codes
        else:
            raise TypeError(f"Cannot compare Origin to type: {type(other)}")

    def __ne__(self, other: object) -> np.ndarray:
        return ~self.__eq__(other)

    def mu(self) -> np.ndarray:
        """
        Return the gravitational parameter of the origin.

        Returns
        -------
        mu : `~numpy.ndarray` (N)
            The gravitational parameter of the origin in au^3 / day^2.

        Raises
        ------
        ValueError
            If the origin code is not recognized.
        """
        mu = np.empty(len(self.code), dtype=np.float64)
        for code in pc.unique(self.code):
            code = code.as_py()
            mask = pc.equal(self.code, code).to_numpy(zero_copy_only=False)
            if code == "SOLAR_SYSTEM_BARYCENTER":
                mu[mask] = OriginGravitationalParameters.SOLAR_SYSTEM_BARYCENTER()
            elif code in OriginGravitationalParameters.__members__:
                mu[mask] = OriginGravitationalParameters[code].value
            else:
                raise ValueError(f"Unknown origin code: {code}")

        return mu



"""
Implementation of Lambert's problem using Izzo's method.

This implementation follows the algorithm described in:
Izzo, D. (2015). Revisiting Lambert's problem. Celestial Mechanics and Dynamical Astronomy, 121(1), 1-15.

Credits: Based on poliastro implementation by Juan Luis Cano Rodríguez and lamberthub by Jorge Martinez
"""

from typing import Tuple, Union

import jax.numpy as jnp
import numpy as np
import numpy.typing as npt
from jax import config, jit, lax, vmap

from ..constants import Constants as C

config.update("jax_enable_x64", True)

MU = C.MU


@jit
def _hyp2f1b(x):
    """Hypergeometric function 2F1(3, 1, 5/2, x), implemented with JAX."""
    # Set a finite number of iterations instead of convergence check
    MAX_ITER = 100

    # Using lax.scan for JAX compatibility
    def body_fun(state, i):
        x, term, res = state

        # Update term and result
        term = term * (3 + i) * (1 + i) / (5 / 2 + i) * x / (i + 1)
        res += term

        return (x, term, res), None

    # Initialize state
    init_state = (
        x.astype(jnp.float64),
        jnp.ones_like(x, dtype=jnp.float64),
        jnp.ones_like(x, dtype=jnp.float64),
    )

    # Run the loop with fixed iterations
    (_, _, res), _ = lax.scan(
        body_fun, init_state, jnp.arange(MAX_ITER, dtype=jnp.float64)
    )

    # Set to infinity for x >= 1
    return jnp.where(x >= 1.0, jnp.inf, res)


@jit
def _compute_y(x, ll):
    """Computes y."""
    return jnp.sqrt(1 - ll**2 * (1 - x**2))


@jit
def _compute_psi(x, y, ll):
    """Computes psi."""
    # Compute the argument for arccos
    arccos_arg = x * y + ll * (1 - x**2)

    # Elliptic case (x < 1.0)
    elliptic = jnp.arccos(jnp.clip(arccos_arg, -1.0, 1.0))

    # Hyperbolic case (x > 1.0)
    hyperbolic = jnp.arcsinh((y - x * ll) * jnp.sqrt(x**2 - 1))

    # Parabolic case (x == 1.0)
    parabolic = jnp.zeros_like(x)

    # Use where statements instead of conditionals
    result = jnp.where(x < 1.0, elliptic, jnp.where(x > 1.0, hyperbolic, parabolic))

    return result


@jit
def _tof_equation_y(x, y, T0, ll, M):
    """Time of flight equation with externally computed y."""

    # Special case for small number of revolutions and specific x range
    # Calculate values for small M case
    eta = y - ll * x
    S_1 = (1 - ll - x * eta) * 0.5
    Q = 4 / 3 * _hyp2f1b(S_1)
    small_M_result = (eta**3 * Q + 4 * ll * eta) * 0.5

    # Calculate values for general case
    psi = _compute_psi(x, y, ll)
    sqrt_term = jnp.sqrt(jnp.abs(1 - x**2))
    numerator = jnp.divide(psi + M * jnp.pi, sqrt_term) - x + ll * y
    denominator = 1 - x**2
    general_result = jnp.divide(numerator, denominator)

    # Use where for conditional selection
    use_small_M = (M == 0) & (jnp.sqrt(0.6) < x) & (x < jnp.sqrt(1.4))
    T_ = jnp.where(use_small_M, small_M_result, general_result)

    result = T_ - T0
    return result


@jit
def _tof_equation(x, T0, ll, M):
    """Time of flight equation."""
    y = _compute_y(x, ll)
    return _tof_equation_y(x, y, T0, ll, M)


@jit
def _tof_equation_p(x, y, T, ll):
    """First derivative of the time of flight equation."""
    return (3 * T * x - 2 + 2 * ll**3 * x / y) / (1 - x**2)


@jit
def _tof_equation_p2(x, y, T, dT, ll):
    """Second derivative of the time of flight equation."""
    return (3 * T + 5 * x * dT + 2 * (1 - ll**2) * ll**3 / y**3) / (1 - x**2)


@jit
def _tof_equation_p3(x, y, _, dT, ddT, ll):
    """Third derivative of the time of flight equation."""
    return (7 * x * ddT + 8 * dT - 6 * (1 - ll**2) * ll**5 * x / y**5) / (1 - x**2)


@jit
def _initial_guess(T, ll, M, low_path):
    """Initial guess for the iterative algorithm."""

    # Single revolution case (M == 0)
    T_0 = jnp.arccos(ll) + ll * jnp.sqrt(1 - ll**2) + M * jnp.pi  # Equation 19
    T_1 = 2 * (1 - ll**3) / 3  # Equation 21

    # Determine initial guess based on T
    x_T0 = (T_0 / T) ** (2 / 3) - 1
    x_T1 = 5 / 2 * T_1 / T * (T_1 - T) / (1 - ll**5) + 1

    # For the middle case, use logarithmic interpolation
    x_middle = jnp.exp(jnp.log(2) * jnp.log(T / T_0) / jnp.log(T_1 / T_0)) - 1

    # Create masks for the cases
    case_T_ge_T0 = T >= T_0
    case_T_lt_T1 = T < T_1

    # Select appropriate single-revolution case
    x_single_rev = jnp.where(
        case_T_ge_T0, x_T0, jnp.where(case_T_lt_T1, x_T1, x_middle)
    )

    # Multiple revolution case
    x_0l = (((M * jnp.pi + jnp.pi) / (8 * T)) ** (2 / 3) - 1) / (
        ((M * jnp.pi + jnp.pi) / (8 * T)) ** (2 / 3) + 1
    )
    x_0r = (((8 * T) / (M * jnp.pi)) ** (2 / 3) - 1) / (
        ((8 * T) / (M * jnp.pi)) ** (2 / 3) + 1
    )

    # Choose high or low path for multi-rev
    x_multi_rev = jnp.where(low_path, jnp.maximum(x_0l, x_0r), jnp.minimum(x_0l, x_0r))

    # Final selection based on M
    result = jnp.where(M == 0, x_single_rev, x_multi_rev)

    return result


@jit
def _householder(p0, T0, ll, M, atol, rtol, maxiter):
    """Find a zero of time of flight equation using the Householder method."""

    def body_fun(state):
        p0, p, iter_count = state

        # Update p0 to be the current p
        p0 = p

        # Compute values needed for the Householder step
        y = _compute_y(p0, ll)
        fval = _tof_equation_y(p0, y, T0, ll, M)
        T = fval + T0
        fder = _tof_equation_p(p0, y, T, ll)
        fder2 = _tof_equation_p2(p0, y, T, fder, ll)
        fder3 = _tof_equation_p3(p0, y, T, fder, fder2, ll)

        # Householder step (quartic)
        numerator = fder**2 - fval * fder2 / 2
        denominator = fder * (fder**2 - fval * fder2) + fder3 * fval**2 / 6

        # Avoid division by zero with a safer approach
        safe_denominator = jnp.where(
            jnp.abs(denominator) < 1e-15, jnp.sign(denominator) * 1e-15, denominator
        )

        # Compute the new value with safeguards against large steps
        delta = fval * (numerator / safe_denominator)
        # Limit step size to prevent divergence
        max_step = jnp.maximum(0.1, jnp.abs(p0))
        delta = jnp.clip(delta, -max_step, max_step)
        p = p0 - delta

        return (p0, p, iter_count + 1)

    def cond_fun(state):
        p0, p, iter_count = state

        # Check both max iterations and convergence criteria
        delta_x = jnp.abs(p - p0)
        converged = delta_x < (rtol * jnp.abs(p0) + atol)
        iter_remaining = iter_count < maxiter

        return (~converged) & iter_remaining

    # Initialize state with different p value to ensure at least one iteration runs
    # Set p to be slightly different from p0 to trigger at least one iteration
    init_p = (
        p0 * 1.1 + 0.01
    )  # Ensure it's different enough to avoid immediate convergence
    init_state = (p0, init_p, 0)

    # Run the Householder iterations
    _, p, iteration_count = lax.while_loop(cond_fun, body_fun, init_state)

    return p


@jit
def _compute_T_min(ll, M, maxiter, atol, rtol):
    """Compute minimum T."""

    # Case 1: ll == 1
    x_T_min_case1 = 0.0
    T_min_case1 = _tof_equation(x_T_min_case1, 0.0, ll, M)

    # Case 2: ll != 1 and M == 0
    x_T_min_case2 = jnp.inf
    T_min_case2 = 0.0

    # Case 3: ll != 1 and M != 0
    # Set x_i > 0 to avoid problems at ll = -1
    x_i = 0.1
    T_i = _tof_equation(x_i, 0.0, ll, M)
    x_T_min_case3 = _halley(x_i, T_i, ll, atol, rtol, maxiter)
    T_min_case3 = _tof_equation(x_T_min_case3, 0.0, ll, M)

    # Select the appropriate case using masks
    is_ll_one = jnp.abs(ll - 1.0) < 1e-10
    is_m_zero = M == 0

    # First choose between case 2 and case 3 based on M
    x_T_min_case23 = jnp.where(is_m_zero, x_T_min_case2, x_T_min_case3)
    T_min_case23 = jnp.where(is_m_zero, T_min_case2, T_min_case3)

    # Then choose between case 1 and the result of case23 based on ll
    x_T_min = jnp.where(is_ll_one, x_T_min_case1, x_T_min_case23)
    T_min = jnp.where(is_ll_one, T_min_case1, T_min_case23)

    return x_T_min, T_min


@jit
def _halley(p0, T0, ll, atol, rtol, maxiter):
    """Find a minimum of time of flight equation using the Halley method."""

    def body_fun(state):
        p0, iter_count = state

        y = _compute_y(p0, ll)
        fder = _tof_equation_p(p0, y, T0, ll)
        fder2 = _tof_equation_p2(p0, y, T0, fder, ll)
        fder3 = _tof_equation_p3(p0, y, T0, fder, fder2, ll)

        # Halley step (cubic)
        p = p0 - 2 * fder * fder2 / (2 * fder2**2 - fder * fder3)

        return (p, iter_count + 1)

    def cond_fun(state):
        p0, iter_count = state
        p, _ = body_fun(state)
        return (jnp.abs(p - p0) >= rtol * jnp.abs(p0) + atol) & (iter_count < maxiter)

    # Run the Halley iterations
    p, _ = lax.while_loop(cond_fun, body_fun, (p0, 0))
    return p


@jit
def _find_xy(ll, T, M, maxiter, atol, rtol, low_path):
    """Computes all x, y for given number of revolutions."""

    # Calculate M_max
    M_max = jnp.floor(T / jnp.pi)
    T_00 = jnp.arccos(jnp.abs(ll)) + jnp.abs(ll) * jnp.sqrt(1 - ll**2)

    # Possibly refine M_max using JAX-native operations
    need_refine = (T < T_00 + M_max * jnp.pi) & (M_max > 0)

    # Define the refine function
    def refine_fn(need_refine):
        _, T_min = _compute_T_min(ll, M_max, maxiter, atol, rtol)
        return jnp.where(T < T_min, M_max - 1, M_max)

    # Use lax.cond for this specific case as it's crucial
    M_max = lax.cond(
        need_refine, lambda _: refine_fn(need_refine), lambda _: M_max, None
    )

    # Check if solution exists - use explicit debug to verify condition
    valid_ll_value = jnp.abs(ll) < 1.0

    valid_M = M <= M_max

    # Get initial guess
    x_0 = _initial_guess(T, ll, M, low_path)

    # Run Householder iterations
    x = _householder(x_0, T, ll, M, atol, rtol, maxiter)
    y = _compute_y(x, ll)

    # Return NaN for invalid cases using jnp.where which is more reliable
    # Check validation condition explicitly again
    x_result = jnp.where(valid_ll_value & valid_M, x, jnp.nan)
    y_result = jnp.where(valid_ll_value & valid_M, y, jnp.nan)

    return x_result, y_result


@jit
def izzo_lambert(
    r1: jnp.ndarray,
    r2: jnp.ndarray,
    tof: float,
    mu: float = MU,
    M: int = 0,
    prograde: bool = True,
    low_path: bool = True,
    maxiter: int = 35,
    atol: float = 1e-10,
    rtol: float = 1e-10,
) -> Tuple[jnp.ndarray, jnp.ndarray]:
    """
    Solves Lambert's problem using Izzo's devised algorithm.

    Parameters
    ----------
    r1: jnp.ndarray
        Initial position vector.
    r2: jnp.ndarray
        Final position vector.
    tof: float
        Time of flight.
    mu: float
        Gravitational parameter, equivalent to GM of attractor body.
    M: int
        Number of revolutions. Must be equal or greater than 0.
    prograde: bool
        If True, specifies prograde motion. Otherwise, retrograde motion is imposed.
    low_path: bool
        If two solutions are available, it selects between high or low path.
    maxiter: int
        Maximum number of iterations.
    atol: float
        Absolute tolerance.
    rtol: float
        Relative tolerance.

    Returns
    -------
    v1: jnp.ndarray
        Initial velocity vector.
    v2: jnp.ndarray
        Final velocity vector.
    """
    # Compute basic geometric quantities
    c = r2 - r1  # Chord
    c_norm = jnp.linalg.norm(c)
    r1_norm = jnp.linalg.norm(r1)
    r2_norm = jnp.linalg.norm(r2)

    # Semiperimeter
    s = (r1_norm + r2_norm + c_norm) * 0.5

    # Normalized vectors
    i_r1 = r1 / r1_norm
    i_r2 = r2 / r2_norm

    # Compute angular momentum unit vector
    i_h = jnp.cross(i_r1, i_r2)
    i_h_norm = jnp.linalg.norm(i_h)
    i_h = i_h / jnp.where(i_h_norm > 0, i_h_norm, 1.0)  # Avoid division by zero

    # Geometry of the problem: lambda parameter
    ll = jnp.sqrt(1 - jnp.minimum(1.0, c_norm / s))

    # Adjust lambda and compute transfer direction based on orbit inclination
    ll_sign = jnp.where(i_h[2] < 0, -1.0, 1.0)
    ll = ll * ll_sign

    # Compute tangential directions
    i_t1 = jnp.where(i_h[2] < 0, jnp.cross(i_r1, i_h), jnp.cross(i_h, i_r1))
    i_t2 = jnp.where(i_h[2] < 0, jnp.cross(i_r2, i_h), jnp.cross(i_h, i_r2))

    # Account for retrograde motion
    ll = jnp.where(~prograde, -ll, ll)
    i_t1 = jnp.where(~prograde, -i_t1, i_t1)
    i_t2 = jnp.where(~prograde, -i_t2, i_t2)

    # Non-dimensional time of flight
    T = jnp.sqrt(2 * mu / s**3) * tof

    # Find x, y using the new approach
    x, y = _find_xy(ll, T, M, maxiter, atol, rtol, low_path)
    # Perform explicit NaN check with JAX operations
    has_nan_x = jnp.isnan(x)
    has_nan_y = jnp.isnan(y)
    has_nans = has_nan_x | has_nan_y

    # Always compute all solution components
    # Reconstruct the solution - always compute these values
    gamma = jnp.sqrt(mu * s / 2)
    rho = (r1_norm - r2_norm) / c_norm
    sigma = jnp.sqrt(1 - rho**2)

    # Compute velocity components
    V_r1 = gamma * ((ll * y - x) - rho * (ll * y + x)) / r1_norm
    V_r2 = -gamma * ((ll * y - x) + rho * (ll * y + x)) / r2_norm
    V_t1 = gamma * sigma * (y + ll * x) / r1_norm
    V_t2 = gamma * sigma * (y + ll * x) / r2_norm

    # Construct velocity vectors
    v1_valid = V_r1 * i_r1 + V_t1 * i_t1
    v2_valid = V_r2 * i_r2 + V_t2 * i_t2

    # Create NaN vectors for invalid cases
    v1_nan = jnp.full_like(r1, jnp.nan)
    v2_nan = jnp.full_like(r2, jnp.nan)

    # Select the appropriate results using where
    v1 = jnp.where(has_nans, v1_nan, v1_valid)
    v2 = jnp.where(has_nans, v2_nan, v2_valid)

    return v1, v2


# Vectorize the Lambert solver
_izzo_lambert_vmap = jit(
    vmap(
        izzo_lambert,
        in_axes=(0, 0, 0, None, None, None, None, None, None, None),
        out_axes=(0, 0),
    )
)


def solve_lambert(
    r1: Union[np.ndarray, jnp.ndarray],
    r2: Union[np.ndarray, jnp.ndarray],
    tof: Union[np.ndarray, float],
    mu: float = MU,
    prograde: bool = True,
    max_iter: int = 35,
    tol: float = 1e-10,
) -> Tuple[np.ndarray, np.ndarray]:
    """
    Solve Lambert's problem for multiple initial and final positions and times of flight.

    This implementation uses Izzo's method which is robust and handles all orbit types.

    Parameters
    ----------
    r1 : array_like (N, 3)
        Initial position vectors in au.
    r2 : array_like (N, 3)
        Final position vectors in au.
    tof : array_like (N) or float
        Times of flight in days.
    mu : float, optional
        Gravitational parameter (GM) of the attracting body in units of au³/day².
    prograde : bool, optional
        If True, assume prograde motion. If False, assume retrograde motion.
    max_iter : int, optional
        Maximum number of iterations for convergence.
    tol : float, optional
        Convergence tolerance.

    Returns
    -------
    v1 : ndarray (N, 3)
        Initial velocity vectors in au/day with origin at the attractor
    v2 : ndarray (N, 3)
        Final velocity vectors in au/day with origin at the attractor
    """
    # Convert inputs to jnp arrays
    r1 = jnp.asarray(r1)
    r2 = jnp.asarray(r2)

    # Handle scalar inputs
    if r1.ndim == 1:
        r1 = r1.reshape(1, -1)
    if r2.ndim == 1:
        r2 = r2.reshape(1, -1)

    # Convert tof to array
    if isinstance(tof, (int, float)):
        tof = jnp.full(r1.shape[0], tof)
    else:
        tof = jnp.asarray(tof)

    # Call vectorized solver (M=0 for single-revolution case)
    v1, v2 = _izzo_lambert_vmap(r1, r2, tof, mu, 0, prograde, True, max_iter, tol, tol)

    # Convert to numpy arrays
    v1 = np.asarray(v1)
    v2 = np.asarray(v2)
    return v1, v2


@jit
def _calculate_c3(
    v1: Union[np.ndarray, jnp.ndarray], body_v: Union[np.ndarray, jnp.ndarray]
) -> jnp.ndarray:
    v_infinity = v1 - body_v

    # Use jnp.linalg.norm for JAX arrays to avoid TracerArrayConversionError
    c3 = jnp.linalg.norm(v_infinity, axis=1) ** 2
    return c3


def calculate_c3(
    v1: Union[np.ndarray, jnp.ndarray], body_v: Union[np.ndarray, jnp.ndarray]
) -> npt.NDArray[np.float64]:
    """
    Calculate the C3 of a spacecraft given its velocity relative to a body.

    Parameters
    ----------
    v1 : array_like (N, 3)
        Velocity of the spacecraft in au/d.
    body_v : array_like (N, 3)
        Velocity of the body in au/d.

    Returns
    -------
    c3 : array_like (N)
        C3 of the spacecraft in au^2/d^2.
    """

    c3 = _calculate_c3(v1, body_v)
    # Convert to numpy array before returning
    c3 = np.asarray(c3)
    return c3



import numpy as np
import pyarrow.compute as pc

from ...coordinates import (
    CartesianCoordinates,
    CoordinateCovariances,
    Origin,
    OriginCodes,
    SphericalCoordinates,
    transform_coordinates,
)
from ...orbits import Orbits, VariantOrbits
from ...propagator import Propagator
from ...time import Timestamp
from ..impacts import (
    CollisionConditions,
    CollisionEvent,
    ImpactMixin,
    ImpactProbabilities,
    calculate_impact_probabilities,
    calculate_impacts,
    calculate_mahalanobis_distance,
)


class MockImpactPropagator(Propagator, ImpactMixin):
    def __getstate__(self):
        state = self.__dict__.copy()
        return state

    def __setstate__(self, state):
        self.__dict__.update(state)

    def _propagate_orbits(self, orbits: Orbits, times: Timestamp) -> Orbits:
        return orbits

    def _detect_collisions(
        self, orbits: Orbits, num_days: float, conditions: CollisionConditions
    ) -> Orbits:
        # Artificially set the orbits.coordinates.times to the end time
        # except for the orbits who impacted
        new_times = orbits.coordinates.time.add_days(num_days)
        orbits = orbits.set_column("coordinates.time", new_times)

        # Do a transform away from the input origin and frame
        orbits_transformed = orbits.set_column(
            "coordinates",
            transform_coordinates(
                orbits.coordinates,
                representation_out=CartesianCoordinates,
                origin_out=OriginCodes.EARTH,
                frame_out="equatorial",
            ),
        )

        # Pick random orbit to impact
        variant = orbits_transformed[0]
        impact = CollisionEvent.from_kwargs(
            orbit_id=variant.orbit_id,
            variant_id=variant.variant_id,
            coordinates=CartesianCoordinates.from_kwargs(
                x=variant.coordinates.x,
                y=variant.coordinates.y,
                z=variant.coordinates.z,
                vx=variant.coordinates.vx,
                vy=variant.coordinates.vy,
                vz=variant.coordinates.vz,
                time=variant.coordinates.time,
                origin=variant.coordinates.origin,
                frame=variant.coordinates.frame,
            ),
            condition_id=["1"],
            collision_coordinates=transform_coordinates(
                variant.coordinates,
                representation_out=SphericalCoordinates,
                origin_out=OriginCodes.EARTH,
                frame_out="itrf93",
            ),
            collision_object=Origin.from_kwargs(code=["EARTH"]),
            stopping_condition=[False],
        )

        return variant, impact


def test_calculate_impacts():
    """
    Tests the i/o of calculate_impacts
    """
    orbits = Orbits.from_kwargs(
        orbit_id=["1", "2", "3"],
        object_id=["1", "2", "3"],
        coordinates=CartesianCoordinates.from_kwargs(
            x=[1.0, 2.0, 3.0],
            y=[1.0, 2.0, 3.0],
            z=[1.0, 2.0, 3.0],
            vx=[1.0, 2.0, 3.0],
            vy=[1.0, 2.0, 3.0],
            vz=[1.0, 2.0, 3.0],
            time=Timestamp.from_iso8601(["2020-01-01T00:00:00"] * 3),
            origin=Origin.from_kwargs(code=["SUN"] * 3),
            frame="ecliptic",
            covariance=CoordinateCovariances.from_sigmas(
                np.array(
                    [
                        [1.0, 1.0, 1.0, 1.0, 1.0, 1.0],
                        [1.0, 1.0, 1.0, 1.0, 1.0, 1.0],
                        [1.0, 1.0, 1.0, 1.0, 1.0, 1.0],
                    ]
                )
            ),
        ),
    )

    prop = MockImpactPropagator()
    variants, impacts = calculate_impacts(
        orbits, 10, prop, num_samples=1000, processes=1, seed=42
    )
    assert pc.all(
        pc.equal(
            variants.coordinates.time.mjd().to_pylist(),
            orbits[0].coordinates.time.add_days(10).mjd().to_pylist()[0],
        )
    )
    assert pc.all(
        pc.equal(
            variants.coordinates.origin.code.unique(),
            orbits.coordinates.origin.code.unique(),
        )
    ).as_py()
    assert variants.coordinates.frame == orbits.coordinates.frame


def test_calculate_impact_probabilities():
    """
    Tests the i/o of calculate_impact_probabilities
    """
    variants = VariantOrbits.from_kwargs(
        orbit_id=["1", "1", "1", "2", "2", "2", "3", "3", "3"],
        object_id=["1", "1", "1", "2", "2", "2", "3", "3", "3"],
        variant_id=["1", "2", "3", "1", "2", "3", "1", "2", "3"],
        coordinates=CartesianCoordinates.from_kwargs(
            x=[1.0, 2.0, 3.0] * 3,
            y=[1.0, 2.0, 3.0] * 3,
            z=[1.0, 2.0, 3.0] * 3,
            vx=[1.0, 2.0, 3.0] * 3,
            vy=[1.0, 2.0, 3.0] * 3,
            vz=[1.0, 2.0, 3.0] * 3,
            time=Timestamp.from_iso8601(["2020-01-01T00:00:00"] * 9),
            origin=Origin.from_kwargs(code=["SUN"] * 9),
            frame="ecliptic",
        ),
    )

    impacts = CollisionEvent.from_kwargs(
        orbit_id=["1", "2", "2"],
        variant_id=["1", "1", "2"],
        coordinates=CartesianCoordinates.from_kwargs(
            x=[1.0, 2.0, 3.0],
            y=[1.0, 2.0, 3.0],
            z=[1.0, 2.0, 3.0],
            vx=[1.0, 2.0, 3.0],
            vy=[1.0, 2.0, 3.0],
            vz=[1.0, 2.0, 3.0],
            time=Timestamp.from_kwargs(
                days=[59200, 59200, 59200],
                nanos=[0, 0, 43200 * 1e9],
                scale="utc",
            ),
            origin=Origin.from_kwargs(code=["SUN"] * 3),
            frame="ecliptic",
        ),
        condition_id=["1", "1", "1"],
        collision_object=Origin.from_kwargs(code=["EARTH", "EARTH", "EARTH"]),
        collision_coordinates=SphericalCoordinates.from_kwargs(
            rho=[1.0, 1.0, 1.0],
            lon=[1.0, 2.0, 3.0],
            lat=[1.0, 2.0, 3.0],
            time=Timestamp.from_kwargs(
                days=[59200, 59200, 59200],
                nanos=[0, 0, 43200 * 1e9],
                scale="utc",
            ),
        ),
        stopping_condition=[False, False, False],
    )

    impact_conditions = CollisionConditions.from_kwargs(
        condition_id=["1"],
        collision_object=Origin.from_kwargs(code=["EARTH"]),
        collision_distance=[1.0],
        stopping_condition=[False],
    )

    ip = calculate_impact_probabilities(variants, impacts, conditions=impact_conditions)

    desired = ImpactProbabilities.from_kwargs(
        condition_id=["1", "1", "1"],
        orbit_id=["1", "2", "3"],
        impacts=[1, 2, 0],
        variants=[3, 3, 3],
        cumulative_probability=[1 / 3, 2 / 3, 0.0],
        mean_impact_time=Timestamp.from_kwargs(
            days=[59200, 59200, None],
            nanos=[0, 43200 * 1e9 / 2, None],
            scale="utc",
            permit_nulls=True,
        ),
        stddev_impact_time=[0.0, 0.25, None],
        minimum_impact_time=Timestamp.from_kwargs(
            days=[59200, 59200, None],
            nanos=[0, 0, None],
            scale="utc",
            permit_nulls=True,
        ),
        maximum_impact_time=Timestamp.from_kwargs(
            days=[59200, 59200, None],
            nanos=[0, 43200 * 1e9, None],
            scale="utc",
            permit_nulls=True,
        ),
    )

    assert ip == desired


def test_calculate_mahalanobis_distance():
    """ """
    observed_orbit = Orbits.from_kwargs(
        orbit_id=["1", "1", "1"],
        object_id=["1", "1", "1"],
        coordinates=CartesianCoordinates.from_kwargs(
            x=[1.0, 2.0, 3.0],
            y=[1.0, 2.0, 3.0],
            z=[1.0, 2.0, 3.0],
            vx=[1.0, 2.0, 3.0],
            vy=[1.0, 2.0, 3.0],
            vz=[1.0, 2.0, 3.0],
            time=Timestamp.from_iso8601(["2020-01-01T00:00:00"] * 3),
            origin=Origin.from_kwargs(code=["SUN"] * 3),
            frame="ecliptic",
            covariance=CoordinateCovariances.from_sigmas(
                np.array(
                    [
                        [1.0, 1.0, 1.0, 1.0, 1.0, 1.0],
                        [1.0, 1.0, 1.0, 1.0, 1.0, 1.0],
                        [1.0, 1.0, 1.0, 1.0, 1.0, 1.0],
                    ]
                )
            ),
        ),
    )

    predicted_orbit = Orbits.from_kwargs(
        orbit_id=["1", "1", "1"],
        object_id=["1", "1", "1"],
        coordinates=CartesianCoordinates.from_kwargs(
            x=[1.0, 2.0, 3.0],
            y=[1.0, 2.0, 3.0],
            z=[1.0, 2.0, 3.0],
            vx=[1.0, 2.0, 3.0],
            vy=[1.0, 2.0, 3.0],
            vz=[1.0, 2.0, 3.0],
            time=Timestamp.from_iso8601(["2020-01-01T00:00:00"] * 3),
            origin=Origin.from_kwargs(code=["SUN"] * 3),
            frame="ecliptic",
        ),
    )

    md = calculate_mahalanobis_distance(observed_orbit, predicted_orbit)
    assert np.all(md == 0.0)

    less_perfect_observed_orbit = Orbits.from_kwargs(
        orbit_id=["1", "1", "1"],
        object_id=["1", "1", "1"],
        coordinates=CartesianCoordinates.from_kwargs(
            x=[1.1, 2.1, 3.1],
            y=[1.1, 2.1, 3.1],
            z=[1.1, 2.1, 3.1],
            vx=[1.1, 2.1, 3.1],
            vy=[1.1, 2.1, 3.1],
            vz=[1.1, 2.1, 3.1],
            time=Timestamp.from_iso8601(["2020-01-01T00:00:00"] * 3),
            origin=Origin.from_kwargs(code=["SUN"] * 3),
            frame="ecliptic",
            covariance=CoordinateCovariances.from_sigmas(
                np.array(
                    [
                        [1.0, 1.0, 1.0, 1.0, 1.0, 1.0],
                        [1.0, 1.0, 1.0, 1.0, 1.0, 1.0],
                        [1.0, 1.0, 1.0, 1.0, 1.0, 1.0],
                    ]
                )
            ),
        ),
    )

    md = calculate_mahalanobis_distance(less_perfect_observed_orbit, predicted_orbit)
    np.testing.assert_almost_equal(md, np.array([0.24494897, 0.24494897, 0.24494897]))



import numpy as np
from numba import jit
from numpy import roots

from adam_core.constants import Constants as c
from adam_core.coordinates import (
    CartesianCoordinates,
    Origin,
    SphericalCoordinates,
    transform_coordinates,
)
from adam_core.orbits import Orbits
from adam_core.time import Timestamp

from .gibbs import calcGibbs
from .herrick_gibbs import calcHerrickGibbs

__all__ = [
    "_calcV",
    "_calcA",
    "_calcB",
    "_calcLambdas",
    "_calcRhos",
    "approxLangrangeCoeffs",
    "calcGauss",
    "gaussIOD",
]

MU = c.MU
C = c.C


@jit("f8(f8[:], f8[:], f8[:])", nopython=True, cache=True)
def _calcV(rho1_hat, rho2_hat, rho3_hat):
    # Vector triple product that gives the area of
    # the "volume of the parallelepiped" or according to
    # to Milani et al. 2008: 3x volume of the pyramid with vertices q, r1, r2, r3.
    # Note that vector triple product rules apply here.
    return np.dot(np.cross(rho1_hat, rho2_hat), rho3_hat)


@jit("f8(f8[:], f8[:], f8[:], f8[:], f8[:], f8, f8, f8)", nopython=True, cache=True)
def _calcA(q1, q2, q3, rho1_hat, rho3_hat, t31, t32, t21):
    # Equation 21 from Milani et al. 2008
    return np.linalg.norm(q2) ** 3 * np.dot(
        np.cross(rho1_hat, rho3_hat), (t32 * q1 - t31 * q2 + t21 * q3)
    )


@jit("f8(f8[:], f8[:], f8[:], f8[:], f8, f8, f8, f8)", nopython=True, cache=True)
def _calcB(q1, q3, rho1_hat, rho3_hat, t31, t32, t21, mu=MU):
    # Equation 19 from Milani et al. 2008
    return (
        mu
        / 6
        * t32
        * t21
        * np.dot(np.cross(rho1_hat, rho3_hat), ((t31 + t32) * q1 + (t31 + t21) * q3))
    )


@jit("UniTuple(f8, 2)(f8, f8, f8, f8, f8)", nopython=True, cache=True)
def _calcLambdas(r2_mag, t31, t32, t21, mu=MU):
    # Equations 16 and 17 from Milani et al. 2008
    lambda1 = t32 / t31 * (1 + mu / (6 * r2_mag**3) * (t31**2 - t32**2))
    lambda3 = t21 / t31 * (1 + mu / (6 * r2_mag**3) * (t31**2 - t21**2))
    return lambda1, lambda3


@jit(
    "UniTuple(f8[:], 3)(f8, f8, f8[:], f8[:], f8[:], f8[:], f8[:], f8[:], f8)",
    nopython=True,
    cache=True,
)
def _calcRhos(lambda1, lambda3, q1, q2, q3, rho1_hat, rho2_hat, rho3_hat, V):
    # This can be derived by taking a series of scalar products of the coplanarity condition equation
    # with cross products of unit vectors in the direction of the observer, in particular, see Chapter 9 in
    # Milani's book on the theory of orbit determination
    numerator = -lambda1 * q1 + q2 - lambda3 * q3
    rho1_mag = np.dot(numerator, np.cross(rho2_hat, rho3_hat)) / (lambda1 * V)
    rho2_mag = np.dot(numerator, np.cross(rho1_hat, rho3_hat)) / V
    rho3_mag = np.dot(numerator, np.cross(rho1_hat, rho2_hat)) / (lambda3 * V)
    rho1 = rho1_mag * rho1_hat
    rho2 = rho2_mag * rho2_hat
    rho3 = rho3_mag * rho3_hat
    return rho1, rho2, rho3


@jit("UniTuple(f8, 2)(f8, f8, f8)", nopython=True, cache=True)
def approxLangrangeCoeffs(r_mag, dt, mu=MU):
    # Calculate the Lagrange coefficients (Gauss f and g series)
    f = 1 - (1 / 2) * mu / r_mag**3 * dt**2
    g = dt - (1 / 6) * mu / r_mag**3 * dt**2
    return f, g


def calcGauss(r1, r2, r3, t1, t2, t3):
    """
    Calculates the velocity vector at the location of the second position vector (r2) with Gauss's
    original method.

    .. math::
        f_1 \approx 1 - \frac{1}{2}\frac{\mu}{r_2^3} (t_1 - t_2)^2

        f_3 \approx 1 - \frac{1}{2}\frac{\mu}{r_2^3} (t_3 - t_2)^2

        g_1 \approx (t_1 - t_2) - \frac{1}{6}\frac{\mu}{r_2^3} (t_1 - t_2)^2

        g_3 \approx (t_3 - t_2) - \frac{1}{6}\frac{\mu}{r_2^3} (t_3 - t_2)^2

        \vec{v}_2 = \frac{1}{f_1 g_3 - f_3 g_1} (-f_3 \vec{r}_1 + f_1 \vec{r}_3)


    For more details on theory see Chapter 5 in Howard D. Curtis' "Orbital Mechanics for
    Engineering Students".

    Parameters
    ----------
    r1 : `~numpy.ndarray` (3)
        Heliocentric position vector at time 1 in cartesian coordinates in units
        of AU.
    r2 : `~numpy.ndarray` (3)
        Heliocentric position vector at time 2 in cartesian coordinates in units
        of AU.
    r3 : `~numpy.ndarray` (3)
        Heliocentric position vector at time 3 in cartesian coordinates in units
        of AU.
    t1 : float
        Time at r1. Units of MJD or JD work or any decimal time format (one day is 1.00) as
        long as all times are given in the same format.
    t2 : float
        Time at r2. Units of MJD or JD work or any decimal time format (one day is 1.00) as
        long as all times are given in the same format.
    t3 : float
        Time at r3. Units of MJD or JD work or any decimal time format (one day is 1.00) as
        long as all times are given in the same format.

    Returns
    -------
    v2 : `~numpy.ndarray` (3)
        Velocity of object at position r2 at time t2 in units of AU per day.
    """
    t12 = t1 - t2
    t32 = t3 - t2
    r2_mag = np.linalg.norm(r2)
    f1, g1 = approxLangrangeCoeffs(r2_mag, t12)
    f3, g3 = approxLangrangeCoeffs(r2_mag, t32)
    v2 = (1 / (f1 * g3 - f3 * g1)) * (-f3 * r1 + f1 * r3)
    return v2


def gaussIOD(
    coords,
    observation_times,
    coords_obs,
    velocity_method="gibbs",
    light_time=True,
    mu=MU,
    max_iter=10,
    tol=1e-15,
):
    """
    Compute up to three intial orbits using three observations in angular equatorial
    coordinates.

    Parameters
    ----------
    coords : `~numpy.ndarray` (3, 2)
        RA and Dec of three observations in units of degrees.
    observation_times : `~numpy.ndarray` (3)
        Times of the three observations in units of decimal days (MJD or JD for example).
    coords_obs : `~numpy.ndarray` (3, 2)
        Heliocentric position vector of the observer at times t in units of AU.
    velocity_method : {'gauss', gibbs', 'herrick+gibbs'}, optional
        Which method to use for calculating the velocity at the second observation.
        [Default = 'gibbs']
    light_time : bool, optional
        Correct for light travel time.
        [Default = True]
    iterate : bool, optional
        Iterate initial orbit using universal anomaly to better approximate the
        Lagrange coefficients.
    mu : float, optional
        Gravitational parameter (GM) of the attracting body in units of
        AU**3 / d**2.
    max_iter : int, optional
        Maximum number of iterations over which to converge to solution.
    tol : float, optional
        Numerical tolerance to which to compute chi using the Newtown-Raphson
        method.

    Returns
    -------
    epochs : `~numpy.ndarry` (<3)
        Epochs in units of decimal days at which preliminary orbits are
        defined.
    orbits : `~numpy.ndarray` ((<3, 6) or (0))
        Up to three preliminary orbits (as cartesian state vectors).
    """
    coords = transform_coordinates(
        SphericalCoordinates.from_kwargs(
            lon=coords[:, 0],
            lat=coords[:, 1],
            origin=Origin.from_kwargs(
                code=np.full(len(coords), "Unknown", dtype="object")
            ),
            frame="equatorial",
        ).to_unit_sphere(),
        representation_out=CartesianCoordinates,
        frame_out="ecliptic",
    )
    rho = coords.values[:, :3]

    rho1_hat = rho[0, :]
    rho2_hat = rho[1, :]
    rho3_hat = rho[2, :]
    q1 = coords_obs[0, :]
    q2 = coords_obs[1, :]
    q3 = coords_obs[2, :]
    q2_mag = np.linalg.norm(q2)

    # Make sure rhohats are unit vectors
    rho1_hat = rho1_hat / np.linalg.norm(rho1_hat)
    rho2_hat = rho2_hat / np.linalg.norm(rho2_hat)
    rho3_hat = rho3_hat / np.linalg.norm(rho3_hat)

    t1 = observation_times[0]
    t2 = observation_times[1]
    t3 = observation_times[2]
    t31 = t3 - t1
    t21 = t2 - t1
    t32 = t3 - t2

    A = _calcA(q1, q2, q3, rho1_hat, rho3_hat, t31, t32, t21)
    B = _calcB(q1, q3, rho1_hat, rho3_hat, t31, t32, t21, mu=mu)
    V = _calcV(rho1_hat, rho2_hat, rho3_hat)
    coseps2 = np.dot(q2, rho2_hat) / q2_mag
    C0 = V * t31 * q2_mag**4 / B
    h0 = -A / B

    if np.isnan(C0) or np.isnan(h0):
        return np.array([])

    # Find roots to eighth order polynomial
    all_roots = roots(
        [
            C0**2,
            0,
            -(q2_mag**2) * (h0**2 + 2 * C0 * h0 * coseps2 + C0**2),
            0,
            0,
            2 * q2_mag**5 * (h0 + C0 * coseps2),
            0,
            0,
            -(q2_mag**8),
        ]
    )

    # Keep only positive real roots (which should at most be 3)
    r2_mags = np.real(all_roots[np.isreal(all_roots) & (np.real(all_roots) >= 0)])
    num_solutions = len(r2_mags)
    if num_solutions == 0:
        return np.array([])

    orbits = []
    epochs = []
    for r2_mag in r2_mags:
        lambda1, lambda3 = _calcLambdas(r2_mag, t31, t32, t21, mu=mu)
        rho1, rho2, rho3 = _calcRhos(
            lambda1, lambda3, q1, q2, q3, rho1_hat, rho2_hat, rho3_hat, V
        )

        if np.dot(rho2, rho2_hat) < 0:
            continue

        # Test if we get the same rho2 as using equation 22 in Milani et al. 2008
        rho2_mag = (h0 - q2_mag**3 / r2_mag**3) * q2_mag / C0
        # np.testing.assert_almost_equal(np.dot(rho2_mag, rho2_hat), rho2)

        r1 = q1 + rho1
        r2 = q2 + rho2
        r3 = q3 + rho3

        if velocity_method == "gauss":
            v2 = calcGauss(r1, r2, r3, t1, t2, t3)
        elif velocity_method == "gibbs":
            v2 = calcGibbs(r1, r2, r3)
        elif velocity_method == "herrick+gibbs":
            v2 = calcHerrickGibbs(r1, r2, r3, t1, t2, t3)
        else:
            raise ValueError(
                "velocity_method should be one of {'gauss', 'gibbs', 'herrick+gibbs'}"
            )

        epoch = t2
        orbit = np.concatenate([r2, v2])

        if light_time:
            rho2_mag = np.linalg.norm(orbit[:3] - q2)
            lt = rho2_mag / C
            epoch -= lt

        if np.linalg.norm(orbit[3:]) >= C:
            continue

        if (np.linalg.norm(orbit[:3]) > 300.0) or (np.linalg.norm(orbit[3:]) > 1.0):
            # Orbits that crash PYOORB:
            # 58366.84446725786 : 9.5544354809296721e+01  1.4093228616761269e+01 -6.6700146960148423e+00 -6.2618123281073522e+01 -9.4167879481188717e+00  4.4421501034359023e+0
            continue

        orbits.append(orbit)
        epochs.append(epoch)

    epochs = np.array(epochs)
    orbits = np.array(orbits)
    if len(orbits) > 0:
        epochs = epochs[~np.isnan(orbits).any(axis=1)]
        orbits = orbits[~np.isnan(orbits).any(axis=1)]

        return Orbits.from_kwargs(
            coordinates=CartesianCoordinates.from_kwargs(
                x=orbits[:, 0],
                y=orbits[:, 1],
                z=orbits[:, 2],
                vx=orbits[:, 3],
                vy=orbits[:, 4],
                vz=orbits[:, 5],
                time=Timestamp.from_mjd(epochs, scale="utc"),
                origin=Origin.from_kwargs(
                    code=np.full(len(orbits), "SUN", dtype="object")
                ),
                frame="ecliptic",
            )
        )

    else:
        return Orbits.empty()



import numpy as np
import pytest

from ...coordinates.cartesian import CartesianCoordinates
from ...coordinates.origin import Origin
from ...observations.exposures import Exposures
from ...observers.observers import Observers
from ...time import Timestamp
from ..bandpasses import get_integrals
from ..magnitude import convert_magnitude, predict_magnitudes


def _delta_mag(template_id: str, source_filter_id: str, target_filter_id: str) -> float:
    ints = get_integrals(
        template_id, np.asarray([source_filter_id, target_filter_id], dtype=object)
    )
    return float(-2.5 * np.log10(float(ints[1]) / float(ints[0])))


def test_convert_magnitude_matches_integral_ratio():
    m_v = np.asarray([20.0, 21.0], dtype=float)
    expected = m_v + _delta_mag("C", "V", "DECam_g")

    out = convert_magnitude(
        m_v,
        np.asarray(["V", "V"], dtype=object),
        np.asarray(["DECam_g", "DECam_g"], dtype=object),
        composition="C",
    )
    assert np.allclose(out, expected, rtol=0.0, atol=1e-12)


def test_convert_magnitude_mix_weights_match_named_template():
    m_v = np.asarray([20.0], dtype=float)
    out_named = convert_magnitude(
        m_v,
        np.asarray(["V"], dtype=object),
        np.asarray(["LSST_r"], dtype=object),
        composition="NEO",
    )
    out_mix = convert_magnitude(
        m_v,
        np.asarray(["V"], dtype=object),
        np.asarray(["LSST_r"], dtype=object),
        composition=(0.5, 0.5),
    )
    assert np.allclose(out_named, out_mix, rtol=0.0, atol=1e-12)


def test_predict_magnitudes_matches_geometry_plus_delta(monkeypatch):
    time = Timestamp.from_mjd([60000, 60000], scale="tdb")
    exposures = Exposures.from_kwargs(
        id=["e1", "e2"],
        start_time=time,
        duration=[30.0, 30.0],
        # Bandpass predictor expects canonical filter IDs (not reported band labels).
        filter=["DECam_g", "LSST_r"],
        observatory_code=["W84", "X05"],
        seeing=[None, None],
        depth_5sigma=[None, None],
    )

    observers = Observers.from_kwargs(
        code=["500", "500"],
        coordinates=CartesianCoordinates.from_kwargs(
            x=[1.0, 1.0],
            y=[0.0, 0.0],
            z=[0.0, 0.0],
            vx=[0.0, 0.0],
            vy=[0.0, 0.0],
            vz=[0.0, 0.0],
            time=time,
            frame="ecliptic",
            origin=Origin.from_kwargs(code=["SUN", "SUN"]),
        ),
    )
    monkeypatch.setattr(exposures, "observers", lambda *args, **kwargs: observers)

    obj = CartesianCoordinates.from_kwargs(
        x=[2.0, 2.0],
        y=[0.0, 0.0],
        z=[0.0, 0.0],
        vx=[0.0, 0.0],
        vy=[0.0, 0.0],
        vz=[0.0, 0.0],
        time=time,
        frame="ecliptic",
        origin=Origin.from_kwargs(code=["SUN", "SUN"]),
    )

    H_v = 15.0
    G = 0.15
    out = predict_magnitudes(
        H_v, obj, exposures, G=G, reference_filter="V", composition="C"
    )

    # Opposition geometry: r=2, delta=1, phase=0 => m_V = H + 5 log10(r*delta)
    m_v = H_v + 5.0 * np.log10(2.0 * 1.0)
    expected = np.asarray(
        [
            m_v + _delta_mag("C", "V", "DECam_g"),
            m_v + _delta_mag("C", "V", "LSST_r"),
        ],
        dtype=float,
    )
    assert np.allclose(out, expected, rtol=0.0, atol=1e-10)


def test_predict_magnitudes_reference_filter_conversion(monkeypatch):
    time = Timestamp.from_mjd([60000], scale="tdb")
    exposures = Exposures.from_kwargs(
        id=["e1"],
        start_time=time,
        duration=[30.0],
        filter=["DECam_g"],
        observatory_code=["W84"],
        seeing=[None],
        depth_5sigma=[None],
    )

    observers = Observers.from_kwargs(
        code=["500"],
        coordinates=CartesianCoordinates.from_kwargs(
            x=[1.0],
            y=[0.0],
            z=[0.0],
            vx=[0.0],
            vy=[0.0],
            vz=[0.0],
            time=time,
            frame="ecliptic",
            origin=Origin.from_kwargs(code=["SUN"]),
        ),
    )
    monkeypatch.setattr(exposures, "observers", lambda *args, **kwargs: observers)

    obj = CartesianCoordinates.from_kwargs(
        x=[2.0],
        y=[0.0],
        z=[0.0],
        vx=[0.0],
        vy=[0.0],
        vz=[0.0],
        time=time,
        frame="ecliptic",
        origin=Origin.from_kwargs(code=["SUN"]),
    )

    H_v = 15.0
    # Convert H_V -> H_LSST_r using the same bandpass delta.
    H_r = H_v + _delta_mag("C", "V", "LSST_r")

    out_from_v = float(
        predict_magnitudes(H_v, obj, exposures, reference_filter="V", composition="C")[
            0
        ]
    )
    out_from_r = float(
        predict_magnitudes(
            H_r, obj, exposures, reference_filter="LSST_r", composition="C"
        )[0]
    )
    assert out_from_r == pytest.approx(out_from_v, abs=1e-10)
